# -*- coding: utf-8 -*-
"""
本模块功能：投资组合的VaR(在险价值)和ES(预期损失)函数包
所属工具包：证券投资分析工具SIAT 
SIAT：Security Investment Analysis Tool
创建日期：2018年6月16日
最新修订日期：2020年7月23日
作者：王德宏 (WANG Dehong, Peter)
作者单位：北京外国语大学国际商学院
作者邮件：wdehong2000@163.com
版权所有：王德宏
用途限制：仅限研究与教学使用，不可商用！商用需要额外授权。
特别声明：作者不对使用本工具进行证券投资导致的任何损益负责！
"""
#==============================================================================
#统一屏蔽一般性警告
import warnings; warnings.filterwarnings("ignore")     
from siat.common import *
from siat.translate import *
#==============================================================================
import matplotlib.pyplot as plt

#处理绘图汉字乱码问题
import sys; czxt=sys.platform
if czxt in ['win32','win64']:
    plt.rcParams['font.sans-serif'] = ['SimHei']  # 设置默认字体
    mpfrc={'font.family': 'SimHei'}

if czxt in ['darwin']: #MacOSX
    plt.rcParams['font.family']= ['Heiti TC']
    mpfrc={'font.family': 'Heiti TC'}

if czxt in ['linux']: #website Jupyter
    plt.rcParams['font.family']= ['Heiti TC']
    mpfrc={'font.family':'Heiti TC'}

# 解决保存图像时'-'显示为方块的问题
plt.rcParams['axes.unicode_minus'] = False 
#==============================================================================
#==============================================================================
def get_portfolio_prices(portfolio,fromdate,todate):
    """
    功能：抓取投资组合portfolio的每日价值
    输入：投资组合portfolio，开始日期，结束日期
    fromdate: 样本开始日期。格式：'YYYY-MM-DD'
    todate: 样本结束日期。既可以是今天日期，也可以是一个历史日期    
    
    输出：投资组合的价格序列，按照日期升序排列
    """
    
    #仅为调试用的函数入口参数，正式使用前需要注释掉！
    """
    portfolio={'Market':('US','^GSPC'),'AAPL':0.5,'MSFT':0.3,'IBM':0.2}
    fromdate='2019-8-1'
    todate  ='2019-8-31'
    """
    #解构投资组合
    _,mktidx,tickerlist,sharelist=decompose_portfolio(portfolio)
    
    #检查股票列表个数与份额列表个数是否一致
    if len(tickerlist) != len(sharelist):
        print("#Error(get_portfolio_prices): numbers of stocks and shares mismatch.")
        return None        
    
    #从雅虎财经抓取股票价格
    import siat.security_prices as ssp
    p=ssp.get_prices_yahoo(tickerlist,fromdate,todate)
    if p is None: return None
    
    import pandas as pd
    #计算投资组合的开盘价
    op=p['Open']
    #计算投资组合的价值
    oprice=pd.DataFrame(op.dot(sharelist))
    oprice.rename(columns={0: 'Open'}, inplace=True)    

    #计算投资组合的收盘价
    cp=p['Close']
    #计算投资组合的价值
    cprice=pd.DataFrame(cp.dot(sharelist))
    cprice.rename(columns={0: 'Close'}, inplace=True) 
    
    #计算投资组合的调整收盘价
    acp=p['Adj Close']
    #计算投资组合的价值
    acprice=pd.DataFrame(acp.dot(sharelist))
    acprice.rename(columns={0: 'Adj Close'}, inplace=True) 
    
    #计算投资组合的交易量
    vol=p['Volume']
    #计算投资组合的价值
    pfvol=pd.DataFrame(vol.dot(sharelist))
    pfvol.rename(columns={0: 'Volume'}, inplace=True) 
    
    #计算投资组合的交易金额
    for t in tickerlist:
        p['Amount',t]=p['Close',t]*p['Volume',t]
    amt=p['Amount']
    #计算投资组合的价值
    pfamt=pd.DataFrame(amt.dot(sharelist))
    pfamt.rename(columns={0: 'Amount'}, inplace=True) 

    #合成开盘价、收盘价、调整收盘价、交易量和交易金额
    pf1=pd.merge(oprice,cprice,how='inner',left_index=True,right_index=True)    
    pf2=pd.merge(pf1,acprice,how='inner',left_index=True,right_index=True)
    pf3=pd.merge(pf2,pfvol,how='inner',left_index=True,right_index=True)
    pf4=pd.merge(pf3,pfamt,how='inner',left_index=True,right_index=True)
    pf4['Ret']=pf4['Close'].pct_change()

    #获得期间的市场收益率：假设无风险收益率非常小，可以忽略
    m=ssp.get_prices_yahoo(mktidx,fromdate,todate)
    m['Mkt']=m['Close'].pct_change()
    rf_df=m[['Mkt']]
    
    #合并pf4与rf_df
    prices=pd.merge(pf4,rf_df,how='left',left_index=True,right_index=True)

    #提取日期和星期几
    prices['Date']=prices.index.strftime("%Y-%m-%d")
    prices['Weekday']=prices.index.weekday+1

    prices['Portfolio']=str(tickerlist)
    prices['Shares']=str(sharelist)
    prices['Adjustment']=prices.apply(lambda x: \
          False if x['Close']==x['Adj Close'] else True, axis=1)

    pfdf=prices[['Portfolio','Shares','Date','Weekday', \
                 'Open','Close','Adj Close','Adjustment', \
                'Volume','Amount','Ret','Mkt']]  

    #判断空值，控制空值可能引起的程序崩溃
    if pfdf is None:
        print("#Error(get_portfolio_prices): failed to retrieve portfolio data")
        return None
    pfdf.dropna(inplace=True)
    if (pfdf is None) or (len(pfdf)==0):
        print("#Error(get_portfolio_prices): failed to retrieve portfolio data")
        return None   

    return pfdf      

    
#==============================================================================
def calc_VaR_normal_standard(ret_series,future_days=1,alpha=0.99):
    
    """
    功能：VaR算法之标准正态法
    输入参数：收益率序列(非百分比)，未来持有时间(天)，置信度
    输出参数：VaR比率，正数
    """
    #去掉空值
    r=ret_series[~ret_series.isnull()]
    
    import numpy as np
    from scipy import stats
    
    z=stats.norm.ppf(1-alpha)
    
    miu_daily=np.mean(r)
    miu_days=np.power(miu_daily+1,future_days)-1
    
    sigma_daily=np.std(r)    
    sigma_days=np.sqrt(future_days)*sigma_daily
    
    VaR_ratio=abs(miu_days+z*sigma_days)

    return VaR_ratio

#==============================================================================
def calc_ES_normal_standard(ret_series,future_days=1,alpha=0.99):
    """
    功能：计算ES，标准正态法
    输入参数：收益率序列，未来持有日期，置信度
    输出：ES比率，正数
    """
    #去掉空值
    r=ret_series[~ret_series.isnull()]
    
    import numpy as np
    from scipy import stats
    
    z=stats.norm.ppf(1-alpha)
    miu_daily=np.mean(r)
    miu_days=np.power(miu_daily+1,future_days)-1
    
    sigma_daily=np.std(r)    
    sigma_days=np.sqrt(future_days)*sigma_daily
    
    zES=-stats.norm.pdf(z)/(1-alpha)
    ratio=abs(miu_days+zES*sigma_days)
    
    return ratio

#==============================================================================
def calc_VaR_normal_modified(ret_series,future_days=1,alpha=0.99):
    """
    功能：VaR算法之修正正态法
    #输入参数：日收益率序列(非百分比)，未来持有日期，置信度
    #输出参数：VaR比率，正数
    """
    #去掉空值
    r=ret_series[~ret_series.isnull()]
    
    from scipy import stats
    import numpy as np

    z=np.abs(stats.norm.ppf(1-alpha))
    S=stats.skew(r)
    K=stats.kurtosis(r)
    
    t1=1/6*(np.power(z,2)-1)*S
    t2=1/24*(np.power(z,3)-3*z)*K
    t3=1/36*(2*np.power(z,3)-5*z)*np.power(S,2)
    t=z+t1+t2-t3
    
    miu_daily=np.mean(r)
    miu_days=np.power(miu_daily+1,future_days)-1
    sigma_daily=np.std(r)    
    sigma_days=np.sqrt(future_days)*sigma_daily
    
    ratio=abs(miu_days+t*sigma_days)
    
    return ratio

#==============================================================================
def calc_ES_normal_modified(ret_series,future_days=1,alpha=0.99):
    """
    功能：ES算法之修正正态法
    #输入参数：日收益率序列(非百分比)，未来持有日期，置信度
    #输出参数：ES比率，正数
    """
    #去掉空值
    r=ret_series[~ret_series.isnull()]
    
    from scipy import stats
    import numpy as np
    
    #计算替代z的t值
    z=np.abs(stats.norm.ppf(1-alpha))
    S=stats.skew(r)
    K=stats.kurtosis(r)
    
    t1=1/6*(np.power(z,2)-1)*S
    t2=1/24*(np.power(z,3)-3*z)*K
    t3=1/36*(2*np.power(z,3)-5*z)*np.power(S,2)
    t=z+t1+t2-t3
    
    miu_daily=np.mean(r)
    miu_days=np.power(miu_daily+1,future_days)-1
    sigma_daily=np.std(r)    
    sigma_days=np.sqrt(future_days)*sigma_daily
    
    #使用t替代原来的z    
    zES=-stats.norm.pdf(t)/(1-alpha)
    ratio=abs(miu_days+zES*sigma_days)
    
    return ratio

#==============================================================================
def get_grouped_rets(ret_series,groupsize=1):
    """
    功能：给定收益率序列pfdf，按照组的大小，在pfdf内创建滚动分组，
    计算每组组内的累计收益率，消除空值，返回组收益率序列
    """
    #测试用数据，测试后应注释掉
    """
    portfolio={'Market':('US','^GSPC'),'AAPL':0.5,'MSFT':0.3,'IBM':0.2}
    fromdate='2019-7-31'
    todate  ='2019-8-31'
    groupsize=2
    #获得投资组合的收益率
    pfdf=get_portfolio_prices(portfolio,fromdate,todate)    
    import pandas as pd
    ret_series=pd.Series(pfdf['Ret'])
    """    

    #定义组内累计收益率计算方法
    cumret=lambda x:(x+1.0).prod()-1.0
    #使用滚动窗口分组，计算各组收益率gret    
    gret_series=ret_series.rolling(groupsize).apply(cumret)
    gret_series=gret_series[~gret_series.isnull()]
    
    #返回组收益率序列
    return gret_series

#==============================================================================
def calc_VaR_historical(ret_series,future_days=1,alpha=0.99):
    """
    功能：VaR算法之历史模拟法
    输入参数：历史日收益率序列，未来持有日期，置信度
    输出：VaR比率，正数
    """
    #去掉空值
    r0=ret_series[~ret_series.isnull()]
    #按未来持有期间数计算分组收益率
    r=get_grouped_rets(r0,future_days)
    
    import numpy as np
    n=len(r)
    t=int(n*(1-alpha))
    SR=np.sort(r)
    
    if t>=1:
        A=SR[t-1]   #SR的第一个元素的序号是0
    else:
        A=SR[0]
    
    VaR_ratio=abs(A)
    
    return VaR_ratio

#==============================================================================
def calc_ES_historical(ret_series,future_days=1,alpha=0.99):
    """
    功能：ES算法之历史模拟法
    输入参数：历史日收益率序列，未来持有日期，置信度
    输出：ES比率，正数
    要求：足够多的历史数据
    """
    #去掉空值
    r0=ret_series[~ret_series.isnull()]
    #按未来持有期间数计算分组收益率
    r=get_grouped_rets(r0,future_days)
    
    import numpy as np
    n=len(r)
    t=int(n*(1-alpha))
    SR=np.sort(r)
    
    if t>2:
        #SR中第t个元素是VaR，第0~(t-1)个元素的均值是ES
        #SR的第一个元素的序号是0
        A=np.mean(SR[0:(t-2)])  
    else:
        A=SR[0]
    
    ratio=abs(A)
    
    return ratio

if __name__ == '__main__':
    Market={'Market':('China','000001.SS')}
    Stocks={'300782.SZ':2,'300661.SZ':3,'688019.SS':4}
    portfolio=dict(Market,**Stocks)
    prices=get_portfolio_prices(portfolio,'2019-7-20','2020-7-20')
    ret_series=prices['Ret']
    future_days=3
    alpha=0.99
    
#==============================================================================
def calc_VaR_montecarlo(ret_series,future_days=1,alpha=0.99, \
                        random=10000,mctype='random'):
    """
    功能：VaR算法之蒙特卡洛模拟法，持有多日
    输入参数：历史日收益率序列，未来持有天数，置信度，重复模拟次数，模拟类型
    注：重复模拟次数越多，准确率就越高，但耗时也越多
    输出：持有多天的VaR(金额)
    """
    #去掉空值
    r=ret_series[~ret_series.isnull()]
    
    import pandas as pd
    import numpy as np
    #蒙特卡洛模拟类型：随机数or超采样
    if mctype=='random':    #随机数产生新的序列
        #取得历史日收益率的均值和标准差
        miu=np.mean(r)
        sigma=np.std(r)
        
        #指定随机数种子
        np.random.seed(12345)
        #按照历史日收益率的均值和标准差重复模拟一定次数，生成新的日收益率序列
        RR=pd.Series(np.random.normal(miu,sigma,random))
    else:   #超采样产生新的序列
        #将收益率序列转变为df
        r1=pd.DataFrame(r)
        r2=r1.sample(n=random,replace=True)
        r2.sort_index(inplace=True)
        RR=pd.Series(r2.iloc[:,0])
        """
        r1=np.random.choice(r,size=random,replace=True)
        RR=pd.Series(r1)
        """
        
    #基于新的日收益率序列，使用标准正态法计算VaR
    ratio=calc_VaR_normal_standard(RR,future_days,alpha)        
    #ratio=calc_VaR_historical(RR,future_days,alpha)
    
    return ratio

if __name__ == '__main__':
    random=1000
    mctype='oversampling'

#==============================================================================
def calc_ES_montecarlo(ret_series,future_days=1,alpha=0.99, \
                        random=10000,mctype='random'):
    """
    功能：ES算法之蒙特卡洛模拟法，持有多日
    输入参数：历史日收益率序列，未来持有天数，置信度，重复模拟次数，模拟类型
    注：重复模拟次数越多，准确率就越高，但耗时也越多
    输出：持有多天的ES比率
    """
    #去掉空值
    r=ret_series[~ret_series.isnull()]
    
    import pandas as pd
    import numpy as np
    #蒙特卡洛模拟类型：随机数or超采样
    if mctype=='random':    #随机数产生新的序列
        #取得历史日收益率的均值和标准差
        miu=np.mean(r)
        sigma=np.std(r)
        
        #指定随机数种子
        np.random.seed(12345)
        #按照历史日收益率的均值和标准差重复模拟一定次数，生成新的日收益率序列
        RR=pd.Series(np.random.normal(miu,sigma,random))
    else:   #超采样产生新的序列
        #将收益率序列转变为df
        r1=pd.DataFrame(r)
        r2=r1.sample(n=random,replace=True)
        r2.sort_index(inplace=True)
        RR=pd.Series(r2.iloc[:,0])
        """
        r1=np.random.choice(r,size=random,replace=True)
        RR=pd.Series(r1)
        """
        
    #基于新的日收益率序列，使用标准正态法计算ES
    ratio=calc_ES_normal_standard(RR,future_days,alpha)        
    
    return ratio

#==============================================================================

def get_VaR_portfolio(portfolio,today,future_days=1,alpha=0.99, \
                      pastyears=1,model='normal_standard',printout=True):
    """
    功能：基于指定模型model，计算投资组合portfolio的VaR金额和比率
    """

    #检查model类型
    modellist=['normal_standard','normal_modified','historical','montecarlo', \
               'mc_oversampling','all']
    modeltyp=model.lower()
    if modeltyp not in modellist:
        print("#Error(get_VaR_portfolio): Unsupported type of model,",model)
        print("Supported models:",modellist)
        return None,None
    
    #获得样本起始日期
    start=get_start_date(today,pastyears)
    if start is None: return None,None
    #抓取投资组合股价和收益率
    prices=get_portfolio_prices(portfolio,start,today)
    #判断空值，控制空值可能引起的程序崩溃
    if prices is None:
        print("#Error(get_VaR_portfolio): failed to retrieve portfolio prices")
        return None,None
    prices.dropna(inplace=True)
    if (prices is None) or (len(prices)==0):
        print("#Error(get_VaR_portfolio): failed to retrieve portfolio prices")
        return None,None        
    
    num=len(prices)
    ret_series=prices['Ret']
    
    #计算当日头寸
    position=round(prices['Close'][-1],2)
    VaRlist=[]; ratiolist=[]
    #标准正态法
    if modeltyp in ['normal_standard','all']:
        ratio=calc_VaR_normal_standard(ret_series,future_days,alpha)
        VaR=-round(position*ratio,2)
        VaRlist=VaRlist+[VaR]
        VaR_ratio=round(ratio,4)
        ratiolist=ratiolist+[VaR_ratio]
    
    #修正正态法
    if modeltyp in ['normal_modified','all']:
        ratio=calc_VaR_normal_modified(ret_series,future_days,alpha)
        VaR=-round(position*ratio,2)
        VaRlist=VaRlist+[VaR]
        VaR_ratio=round(ratio,4)
        ratiolist=ratiolist+[VaR_ratio]
    
    #历史模拟法
    if modeltyp in ['historical','all']:
        ratio=calc_VaR_historical(ret_series,future_days,alpha)
        VaR=-round(position*ratio,2)
        VaRlist=VaRlist+[VaR]
        VaR_ratio=round(ratio,4)
        ratiolist=ratiolist+[VaR_ratio]    
    
    #蒙特卡洛模拟法，随机数，默认
    if modeltyp in ['montecarlo','all']:
        ratio=calc_VaR_montecarlo(ret_series,future_days,alpha,mctype='random')
        VaR=-round(position*ratio,2)
        VaRlist=VaRlist+[VaR]
        VaR_ratio=round(ratio,4)
        ratiolist=ratiolist+[VaR_ratio]
    
    #蒙特卡洛模拟法，超采样
    if modeltyp in ['mc_oversampling','all']:
        ratio=calc_VaR_montecarlo(ret_series,future_days,alpha, \
                                  mctype='mc_oversampling')
        VaR=-round(position*ratio,2)
        VaRlist=VaRlist+[VaR]
        VaR_ratio=round(ratio,4)
        ratiolist=ratiolist+[VaR_ratio]    
    
    if not printout: return VaRlist,ratiolist
    
    #输出VaR金额和比率
    print("\n===== 投资组合的在险价值VaR =====")
    _,_,tickerlist,sharelist=decompose_portfolio(portfolio)
    print("成分股列表  :",tickerlist)
    print("成分股的配置:",sharelist)
    print("当前日期    :",today)
    print("当前头寸    :",format(position,','))
    print("预计持有天数:",future_days)
    print("置信度水平  :",str(int(alpha*100))+'%')
    print("历史样本年数:",pastyears)
    print("样本数量    :",num)
    
    import datetime as dt; today=dt.date.today()
    footnote="*数据来源：雅虎财经，"+str(today)
    if not (modeltyp == 'all'):
        print("使用的模型  :",model)
        print("VaR金额/比率:",format(VaR,','), \
              '\b，',str(round(VaR_ratio*100,2))+'%')
        
        print(footnote)
        return VaR,VaR_ratio
    
    print("\nVaR金额/比率：")
    modellist.pop()
    for m in modellist:
        pos=modellist.index(m)
        v=VaRlist[pos]
        r=ratiolist[pos]
        print("  ",format(v,','),'\b,', \
              str(round(r*100,2))+'%'+' ('+m+')')

    print(footnote)

    return VaRlist,ratiolist

if __name__ == '__main__':
    portfolio={'Market':('China','000001.SS'),'000661.SZ':1,'603392.SS':2, \
               '300601.SZ':3}
    today='2020-7-20'
    future_days=1
    alpha=0.99
    pastyears=1
    model='all'
    printout=True

#==============================================================================
def get_ES_portfolio(portfolio,today,future_days=1,alpha=0.99, \
                      pastyears=1,model='normal_standard',printout=True):
    """
    功能：基于指定模型model，计算投资组合portfolio的ES金额和比率
    """

    #检查model类型
    modellist=['normal_standard','normal_modified','historical','montecarlo', \
               'mc_oversampling','all']
    modeltyp=model.lower()
    if modeltyp not in modellist:
        print("#Error(get_ES_portfolio): Unsupported type of model,",model)
        print("Supported models:",modellist)
        return None,None
    
    #获得样本起始日期
    start=get_start_date(today,pastyears)
    if start is None: return None,None
    #抓取投资组合股价和收益率
    prices=get_portfolio_prices(portfolio,start,today)
    #判断空值，控制空值可能引起的程序崩溃
    if prices is None:
        print("#Error(get_ES_portfolio): failed to retrieve portfolio prices")
        return None,None
    prices.dropna(inplace=True)
    if (prices is None) or (len(prices)==0):
        print("#Error(get_ES_portfolio): failed to retrieve portfolio prices")
        return None,None        
    
    num=len(prices)
    ret_series=prices['Ret']
    
    #计算当日头寸
    position=round(prices['Close'][-1],2)
    ESlist=[]; ratiolist=[]
    #标准正态法
    if modeltyp in ['normal_standard','all']:
        ratio=calc_ES_normal_standard(ret_series,future_days,alpha)
        ES=-round(position*ratio,2)
        ESlist=ESlist+[ES]
        ES_ratio=round(ratio,4)
        ratiolist=ratiolist+[ES_ratio]
    
    #修正正态法
    if modeltyp in ['normal_modified','all']:
        ratio=calc_ES_normal_modified(ret_series,future_days,alpha)
        ES=-round(position*ratio,2)
        ESlist=ESlist+[ES]
        ES_ratio=round(ratio,4)
        ratiolist=ratiolist+[ES_ratio]
    
    #历史模拟法
    if modeltyp in ['historical','all']:
        ratio=calc_ES_historical(ret_series,future_days,alpha)
        ES=-round(position*ratio,2)
        ESlist=ESlist+[ES]
        ES_ratio=round(ratio,4)
        ratiolist=ratiolist+[ES_ratio]    
    
    #蒙特卡洛模拟法，随机数，默认
    if modeltyp in ['montecarlo','all']:
        ratio=calc_ES_montecarlo(ret_series,future_days,alpha,mctype='random')
        ES=-round(position*ratio,2)
        ESlist=ESlist+[ES]
        ES_ratio=round(ratio,4)
        ratiolist=ratiolist+[ES_ratio]
    
    #蒙特卡洛模拟法，超采样
    if modeltyp in ['mc_oversampling','all']:
        ratio=calc_ES_montecarlo(ret_series,future_days,alpha, \
                                  mctype='mc_oversampling')
        ES=-round(position*ratio,2)
        ESlist=ESlist+[ES]
        ES_ratio=round(ratio,4)
        ratiolist=ratiolist+[ES_ratio]    
    
    if not printout: return ESlist,ratiolist
    
    #输出ES金额和比率
    print("\n===== 投资组合的预期不足ES =====")
    _,_,tickerlist,sharelist=decompose_portfolio(portfolio)
    print("成分股列表  :",tickerlist)
    print("成分股配置  :",sharelist)
    print("当前日期    :",today)
    print("当前头寸    :",format(position,','))
    print("预计持有天数:",future_days)
    print("置信度水平  :",str(int(alpha*100))+'%')
    print("历史样本年数:",pastyears)
    print("样本个数    :",num)

    import datetime as dt; today=dt.date.today()
    footnote="*数据来源：雅虎财经，"+str(today)
    
    if not (modeltyp == 'all'):
        print("使用的模型  :",model)
        print("ES金额/比率 ：",format(ES,','), \
              '\b,',str(round(ES_ratio*100,2))+'%')
        print(footnote)
        return ES,ES_ratio
    
    print("ES金额/比率 ：")
    modellist.pop()
    for m in modellist:
        pos=modellist.index(m)
        v=ESlist[pos]
        r=ratiolist[pos]
        print("  ",format(v,','),'\b,', \
              str(round(r*100,2))+'%'+' ('+m+')')

    print(footnote)

    return ESlist,ratiolist

if __name__ == '__main__':
    portfolio={'Market':('China','000001.SS'),'000661.SZ':1,'603392.SS':2, \
               '300601.SZ':3}
    today='2020-7-20'
    future_days=1
    alpha=0.99
    pastyears=1
    model='all'
    printout=True

#===========================================================================
def ret_Normality_SW(ret_series,siglevel=0.05):
    """
    功能：测试一个投资组合portfolio在给定期间内（fromdate,todate）的收益率序列
    是否符合正态分布
    输入参数：投资组合，开始日期，结束日期，显著性要求水平
    输出：收益率序列正态性检验的W, p-value, Skewness, Kurtosis
    【Shapiro-Wilk正态性检验】原假设：服从正态分布
    """
    #去掉空值
    ret_series=ret_series[~ret_series.isnull()]

    from scipy import stats
    (W,p_value)=stats.shapiro(ret_series)
    
    S=stats.skew(ret_series)
    K=stats.kurtosis(ret_series)
    
    return round(W,4),round(p_value,4),round(S,2),round(K,2)

#===========================================================================
def portfolio_ret_Normality_SW(portfolio,fromdate,todate, \
                               siglevel=0.05,printout=True):
    """
    功能：测试一个投资组合portfolio在给定期间内（fromdate,todate）的收益率序列
    是否符合正态分布
    输入参数：投资组合，开始日期，结束日期，显著性要求水平
    输出：收益率序列正态性检验的p-value
    【Shapiro-Wilk正态性检验】原假设：服从正态分布
    """
    #抓取投资组合股价和收益率
    prices=get_portfolio_prices(portfolio,fromdate,todate)
    if prices is None: return None,None
    ret_series=prices['Ret']
    num=len(prices)-1
    
    #检验正态分布
    W,p_value,S,K=ret_Normality_SW(ret_series,siglevel)    
    
    if not printout: return W,p_value,S,K
    
    _,_,tickerlist,sharelist=decompose_portfolio(portfolio)
    print("*** Shapiro-Wilk Normality Test ***")
    print("    Stock(s):",tickerlist)
    print("    Holding proportion:",sharelist)    
    print("    Sampling period   :",fromdate,'to',todate)
    print("    Observations      :",num)
    print("    Null hypothesis   : Normal")
    print("    W statistic       :",round(W,4))
    print("    p-value           :",round(p_value,4))
    print("    Skewness          :",round(S,2))
    print("    Kurtosis          :",round(K,2))
    if p_value >= siglevel:
        print("Result: Accept null hypothesis, normal")
    else:
        print("Result: Reject null hypothesis, not normal")
    
    footnote="... Data source: Yahoo Finance"
    print(footnote)
    
    return W,p_value,S,K

#==============================================================================
def portfolio_rets_curve(portfolio,fromdate,todate):
    """
    功能：绘制投资组合portfolio在给定期间（fromdate,todate）收益率分布的曲线，
    并于相应的正态分布图对照
    显示：收益率分布的直方图(实线)，相应的正态分布图(虚线)
    x轴为收益率(非百分比)，y轴为频度(Frequency)
    """
    #抓取投资组合股价和收益率
    prices=get_portfolio_prices(portfolio,fromdate,todate)
    if prices is None: return None,None
    
    num=len(prices)
    rets=prices['Ret']
    W,p_value,S,K=ret_Normality_SW(rets)
    _,_,tickerlist,sharelist=decompose_portfolio(portfolio)
    #计算收益率的均值和标准差
    mu=rets.mean()
    sigma=rets.std()

    #生成符合正态分布的随机数，符合股票收益率的均值和标准差
    import numpy as np
    x=mu+sigma*np.random.randn(10000)    
            
    plt.figure(figsize=(8,4))
    import seaborn as sns
    #绘制曲线：股票收益率
    sns.kdeplot(data=rets,shade=True,color='blue',legend=True,label='投资组合',lw=4) 
    #绘制曲线：对应的正态分布
    sns.kdeplot(data=x,shade=True,color='r',legend=True,label='正态分布线',ls='--')
    #设置标题、图例、坐标轴标签
    #plt.ylabel('Frequency')
    
    footnote1='日收益率%'
    footnote2='\nShapiro-Wilk检验：W值'+str(W)+'，p值'+str(p_value)
    footnote3="\n偏度"+str(S)+"，峰度"+str(K)
    
    import datetime as dt; today=dt.date.today()
    footnote4="\n*数据来源：雅虎财经，"+str(today)
    
    footnote=footnote1+footnote2+footnote3+footnote4
    plt.xlabel(footnote)
    
    plt.legend(loc='best')
    titletxt1="投资组合收益率的正态分布检验"
    titletxt2="\n成分股"+str(tickerlist)
    titletxt3="\n持仓配置"+str(sharelist)
    titletxt4="\n样本期间："+str(fromdate)+'至'+str(todate)+ \
                '，样本数量'+str(num)+'个'
    titletxt=titletxt1+titletxt2+titletxt3+titletxt4
    plt.title(titletxt)
    plt.show()

    return
    

#==============================================================================



# -*- coding: utf-8 -*-
"""
本模块功能：证券资产的流动性风险与溢价
所属工具包：证券投资分析工具SIAT 
SIAT：Security Investment Analysis Tool
创建日期：2019年6月18日
最新修订日期：2020年7月16日
作者：王德宏 (WANG Dehong, Peter)
作者单位：北京外国语大学国际商学院
版权所有：王德宏
用途限制：仅限研究与教学使用，不可商用！商用需要额外授权。
特别声明：作者不对使用本工具进行证券投资导致的任何损益负责！
"""

#==============================================================================
#关闭所有警告
import warnings; warnings.filterwarnings('ignore')
from siat.common import *
from siat.translate import *
from siat.grafix import *
from siat.security_prices import *
#==============================================================================
import matplotlib.pyplot as plt

#处理绘图汉字乱码问题
import sys; czxt=sys.platform
if czxt in ['win32','win64']:
    plt.rcParams['font.sans-serif'] = ['SimHei']  # 设置默认字体
    mpfrc={'font.family': 'SimHei'}

if czxt in ['darwin']: #MacOSX
    plt.rcParams['font.family']= ['Heiti TC']
    mpfrc={'font.family': 'Heiti TC'}

if czxt in ['linux']: #website Jupyter
    plt.rcParams['font.family']= ['Heiti TC']
    mpfrc={'font.family':'Heiti TC'}

# 解决保存图像时'-'显示为方块的问题
plt.rcParams['axes.unicode_minus'] = False 
#==============================================================================

def calc_roll_spread(pfdf):
    """
    功能：从给定的股票或投资组合portfolio的数据集df中按期间计算罗尔价差
    投资组合的结构：{'Market':('US','^GSPC'),'AAPL':0.5,'MSFT':0.3,'IBM':0.2}
    输入：投资组合价格df，单一数据集有利于计算滚动指数
    输出：罗尔价差%
    注意：不包括爬虫部分
    """
    sp=pfdf.copy()
    #计算价格序列的价差
    sp['dP']=sp['Close'].diff()
    sp['dP_1']=sp['dP'].shift(1)
    sp2=sp[['Close','dP','dP_1']].copy()      
    sp2.dropna(inplace=True)
    if len(sp2) == 0: return None

    #计算指标，注意cov函数的结果是一个矩阵
    import numpy as np
    rs_cov=abs(np.cov(sp2['dP'],sp2['dP_1'])[0,1])    
    #计算价格均值
    p_mean=sp2['Close'].mean()    
    rs=2*np.sqrt(rs_cov)/p_mean 

    rs_pct=round(rs*100.0,4)
    return rs_pct

if __name__=='__main__':
    pf={'Market':('US','^GSPC'),'AAPL':1.0}
    start='2020-1-1'; end='2020-3-31'
    pfdf=get_portfolio_prices(pf,start,end)
    rs=calc_roll_spread(pfdf)

#==============================================================================    
def roll_spread_portfolio(portfolio,start,end,printout=True):
    """
    功能：按期间计算一个投资组合的罗尔价差，并输出结果
    投资组合的结构：{'Market':('US','^GSPC'),'AAPL':0.5,'MSFT':0.3,'IBM':0.2}
    输入：投资组合，开始日期，结束日期，是否打印结果(默认打印)
    输出：罗尔价差%
    注意：含爬虫部分，调用其他函数
    """

    #仅为测试用
    #portfolio={'Market':('US','^GSPC'),'EDU':0.7,'TAL':0.3}
    #start='2019-06-01'
    #end  ='2019-06-30'

    #检查日期和期间的合理性
    flag,start2,end2=check_period(start,end)
    if not flag:
        print("#Error(roll_spread_portfolio): invalid period for,", start, end)
        return None
    
    #抓取股票价格，构建投资组合价格
    sp=get_portfolio_prices(portfolio,start2,end2)
    if sp is None:
        print("#Error(roll_spread_portfolio): portfolio info not available.")
        return None    

    #计算罗尔价差指标
    rs_pct=calc_roll_spread(sp)
    
    #打印报告
    if printout == True:
        date_start=str(sp.index[0].year)+'-'+str(sp.index[0].month)+ \
            '-'+str(sp.index[0].day)
        date_end=str(sp.index[-1].year)+'-'+str(sp.index[-1].month)+ \
            '-'+str(sp.index[-1].day)            
        print("\n===== 投资组合的流动性风险 =====")
        print("投资组合:",portfolio)
        print("计算期间:",date_start,"to",date_end, \
                  "(可用日期)")
        print("罗尔价差%:",rs_pct)
    
    return rs_pct

if __name__=='__main__':
    pf_aapl={'Market':('US','^GSPC'),'AAPL':1.0}
    rs_aapl=roll_spread_portfolio(pf_aapl,'2019-01-01','2019-01-31')    
    
    pfA={'Market':('US','^GSPC'),'AAPL':0.5,'MSFT':0.3,'IBM':0.2}
    rsA=roll_spread_portfolio(pfA,'2019-01-01','2019-01-31')    
    
#==============================================================================        
def calc_amihud_illiquidity(pfdf):
    """
    功能：从给定的投资组合pfdf中计算阿米胡德非流动性
    投资组合的结构：{'Market':('US','^GSPC'),'AAPL':0.5,'MSFT':0.3,'IBM':0.2}
    输入：投资组合价格pfdf
    输出：阿米胡德非流动性
    注意：不包括爬虫部分
    """
    sp=pfdf.copy()
    #计算阿米胡德非流动性
    sp2=sp[['Ret%','Amount']]
    sp2=sp2.dropna()
    if len(sp2) == 0: return None
    
    import numpy as np
    sp2['Ret/Amt']=np.abs(sp2['Ret%'])/np.log10(sp2['Amount'])
    amihud=round(sp2['Ret/Amt'].mean(),4)
    
    return amihud

if __name__=='__main__':
    pf={'Market':('US','^GSPC'),'AAPL':1.0}
    start='2020-1-1'; end='2020-3-31'
    pfdf=get_portfolio_prices(pf,start,end)
    amihud=calc_amihud_illiquidity(pfdf)

#==============================================================================    
def amihud_illiquidity_portfolio(portfolio,start,end,printout=True):
    """
    功能：按天计算一个投资组合的阿米胡德非流动性指数
    投资组合的结构：{'Market':('US','^GSPC'),'AAPL':0.5,'MSFT':0.3,'IBM':0.2}
    输入：投资组合，开始日期，结束日期，是否打印结果
    输出：阿米胡德非流动性指数
    注意：含爬虫部分，调用其他函数
    """

    #仅为测试用
    #portfolio={'Market':('US','^GSPC'),'EDU':0.7,'TAL':0.3}
    #start='2019-06-01'
    #end  ='2019-06-30'

    #检查日期和期间的合理性
    flag,start2,end2=check_period(start,end)
    if not flag:
        print("#Error(amihud_illiquidity_portfolio): invalid period for,", start, end)
        return None

    #抓取股票价格，构建投资组合价格
    sp=get_portfolio_prices(portfolio,start2,end2)

    #计算指标
    amihud=calc_amihud_illiquidity(sp)
    
    #打印报告
    if printout == True:
        date_start=str(sp.index[0].year)+'-'+str(sp.index[0].month)+ \
            '-'+str(sp.index[0].day)
        date_end=str(sp.index[-1].year)+'-'+str(sp.index[-1].month)+ \
            '-'+str(sp.index[-1].day)            
        print("\n===== 投资组合的流动性风险 =====")
        print("投资组合:",portfolio)
        print("计算期间:",date_start,"至",date_end, \
                  "(可用日期)")
        print("阿米胡德非流动性:",amihud,"(对数算法)")
    
    return amihud

if __name__=='__main__':
    pf_aapl={'Market':('US','^GSPC'),'AAPL':1.0}
    amihud_aapl=amihud_illiquidity_portfolio(pf_aapl,'2019-01-01','2019-01-31')    

#==============================================================================        
def calc_ps_liquidity(pfdf):
    """
    功能：从给定的投资组合pfdf中计算帕斯托-斯坦堡流动性，原始公式
    投资组合的结构：{'Market':('US','^GSPC'),'AAPL':0.5,'MSFT':0.3,'IBM':0.2}
    输入：投资组合价格df
    输出：帕斯托-斯坦堡流动性
    注意：不包括爬虫部分
    """
    reg=pfdf.copy()
    
    reg['Ret-RF']=reg['Ret%']-reg['RF']
    reg['Ret-RF_1']=reg['Ret-RF'].shift(1)
    reg['Mkt']=reg['Mkt-RF']+reg['RF']
    reg['Mkt_1']=reg['Mkt'].shift(1)
    reg['Amount_1']=reg['Amount'].shift(1)
    
    import numpy as np
    reg1=reg[['Ret-RF','Mkt_1','Ret-RF_1','Amount_1']]
    reg1=reg1.dropna()
    if len(reg1) == 0: return None
    reg1['signAmt_1']=np.sign(reg1['Ret-RF_1'])*np.log10(reg1['Amount_1'])
    reg2=reg1[['Ret-RF','Mkt_1','signAmt_1']].copy()
    
    #回归前彻底删除带有NaN和inf等无效值的样本，否则回归中可能出错
    reg2=reg2[~reg2.isin([np.nan, np.inf, -np.inf]).any(1)].dropna()
    if len(reg2) == 0: return None
    
    ##计算帕斯托-斯坦堡流动性PSL
    import statsmodels.api as sm
    y=reg2['Ret-RF']
    X=reg2[['Mkt_1','signAmt_1']]
    X1=sm.add_constant(X)
    results=sm.OLS(y,X1).fit() 
    try:
        [alpha,beta,psl]=results.params
    except:
        print("#Error(calc_ps_liquidity): failed to extract reg parms",results.params)
        return None
    
    return round(psl,4)

if __name__=='__main__':
    pf={'Market':('US','^GSPC'),'AAPL':1.0}
    start='2020-1-1'; end='2020-3-31'
    pfdf=get_portfolio_prices(pf,start,end)
    psl=calc_ps_liquidity(pfdf)

#==============================================================================        
def calc_ps_liquidity_modified(pfdf):
    """
    功能：从给定的投资组合pfdf中计算修正的帕斯托-斯坦堡流动性
    投资组合的结构：{'Market':('US','^GSPC'),'AAPL':0.5,'MSFT':0.3,'IBM':0.2}
    输入：投资组合价格df
    输出：修正的帕斯托-斯坦堡流动性，符号内含
    注意：不包括爬虫部分
    """
    reg=pfdf.copy()
    
    reg['Ret-RF']=reg['Ret%']-reg['RF']
    reg['Ret-RF_1']=reg['Ret-RF'].shift(1)
    reg['Mkt']=reg['Mkt-RF']+reg['RF']
    reg['Mkt_1']=reg['Mkt'].shift(1)
    reg['Amount_1']=reg['Amount'].shift(1)
    
    import numpy as np
    reg1=reg[['Ret-RF','Mkt_1','Ret-RF_1','Amount_1']]
    reg1=reg1.dropna()
    if len(reg1) == 0: return None
    
    #修正psl，符号内含
    #reg1['signAmt_1']=np.sign(reg1['Ret-RF_1'])*np.log10(reg1['Amount_1'])
    reg1['signAmt_1']=np.log10(reg1['Amount_1'])
    reg2=reg1[['Ret-RF','Mkt_1','signAmt_1']].copy()
    
    #回归前彻底删除带有NaN和inf等无效值的样本，否则回归中可能出错
    reg2=reg2[~reg2.isin([np.nan, np.inf, -np.inf]).any(1)].dropna()
    if len(reg2) == 0: return None
    
    ##计算帕斯托-斯坦堡流动性PSL
    import statsmodels.api as sm
    y=reg2['Ret-RF']
    X=reg2[['Mkt_1','signAmt_1']]
    X1=sm.add_constant(X)
    results=sm.OLS(y,X1).fit() 
    [alpha,beta,psl]=results.params
    
    return round(psl,4)

if __name__=='__main__':
    pf={'Market':('US','^GSPC'),'AAPL':1.0}
    start='2020-1-1'; end='2020-3-31'
    pfdf=get_portfolio_prices(pf,start,end)
    psl=calc_ps_liquidity_modified(pfdf)

#==============================================================================    
def ps_liquidity_portfolio(portfolio,start,end,printout=True):
    """
    功能：按天计算一个投资组合的帕斯托-斯坦堡流动性
    投资组合的结构：{'Market':('US','^GSPC'),'AAPL':0.5,'MSFT':0.3,'IBM':0.2}
    输入：投资组合，开始日期，结束日期，是否打印结果
    输出：帕斯托-斯坦堡流动性
    注意：含爬虫部分，调用其他函数
    """

    #仅为测试用
    #portfolio={'Market':('US','^GSPC'),'EDU':0.7,'TAL':0.3}
    #start='2019-06-01'
    #end  ='2019-06-30'

    #检查日期和期间的合理性
    flag,start2,end2=check_period(start,end)
    if not flag:
        print("#Error(amihud_illiquidity_portfolio): invalid period for,", start, end)
        return None

    #抓取股票价格，构建投资组合价格
    sp=get_portfolio_prices(portfolio,start2,end2)

    #计算帕斯托-斯坦堡流动性
    psl=calc_ps_liquidity(sp)
    
    #打印报告
    if printout == True:
        date_start=str(sp.index[0].year)+'-'+str(sp.index[0].month)+ \
            '-'+str(sp.index[0].day)
        date_end=str(sp.index[-1].year)+'-'+str(sp.index[-1].month)+ \
            '-'+str(sp.index[-1].day)            
        print("\n===== 投资组合的流动性风险 =====")
        print("投资组合:",portfolio)
        print("计算期间:",date_start,"to",date_end, \
                  "(可用日期)")
        print("Pastor-Stambaugh流动性:",psl,"(对数算法)")
    
    return psl

if __name__=='__main__':
    pf_aapl={'Market':('US','^GSPC'),'AAPL':1.0}
    psl_aapl=ps_liquidity_portfolio(pf_aapl,'2019-01-01','2019-01-31')    

    
#==============================================================================
#==============================================================================
def plot_liquidity_monthly(portfolio,start,end,liquidity_type):
    """
    功能：将资产流动性指标逐月绘图对比
    输入：投资组合，开始/结束日期，流动性指标类别
    输出：流动性指标的逐月数据框
    显示：按月绘图投资组合的流动性指标
    """
    #仅为测试用
    #portfolio={'Market':('US','^GSPC'),'JD':0.3,'BABA':0.7}
    #start='2019-01-01'
    #end='2019-03-31'
    #liquidity_type='roll_spread'    

    #检查日期和期间的合理性
    flag,start2,end2=check_period(start,end)
    if not flag:
        print("#Error(plot_liquidity_monthly): invalid period for,", start, end)
        return None

    #检查：支持的liquidity_type
    liquidity_list=['roll_spread','amihud_illiquidity','ps_liquidity']
    if liquidity_type not in liquidity_list:
        print("#Error(plot_liquidity_monthly): not supported liquidity type")
        print("Supported liquidity type:",liquidity_list)              
        return None         

    #抓取投资组合信息
    print("\n... Searching for portfolio information ...")
    df=get_portfolio_prices(portfolio,start,end)

    #拆分start/end之间的各个年份和月份
    mdlist=calc_monthly_date_range(start,end)
    if len(mdlist) == 0:
        print("#Error(plot_liquidity_monthly): start/end dates inappropriate")
        return None          

    #用于保存流动性指标
    import pandas as pd
    print("... Calculating monthly",liquidity_type,"...")
    rarfunc='calc_'+liquidity_type
    rars=pd.DataFrame(columns=('YM','rar'))
    zeroline=False
    for i in range(0,len(mdlist)):
        startym=mdlist[i][0]
        YM=startym.strftime("%Y-%m")
        #print(YM,end=' ')
        endym=mdlist[i][1]
        pfdf=sample_selection(df,startym,endym)
        rar=eval(rarfunc)(pfdf)
        if rar is None: continue
        
        if rar < 0: zeroline=True
        row=pd.Series({'YM':YM,'rar':rar})
        rars=rars.append(row,ignore_index=True)         
    #print("completed.")
    rars.set_index('YM',inplace=True)    
    
    #绘图
    colname='rar'
    collabel=ectranslate(liquidity_type)
    ylabeltxt=ectranslate(liquidity_type)
    titletxt="证券流动性风险的月度指标"
    
    _,_,tickerlist,sharelist=decompose_portfolio(portfolio)
    if len(tickerlist)==1:
        product=str(tickerlist)
    else:
        product=str(tickerlist)+' in '+str(sharelist)
    import datetime as dt; today=dt.date.today()
    footnote="证券="+product+"\n数据来源：雅虎财经, "+str(today)
    datatag=False
    power=4
    plot_line(rars,colname,collabel,ylabeltxt,titletxt,footnote,datatag, \
              power,zeroline)

    return rars


if __name__=='__main__':
    portfolio={'Market':('US','^GSPC'),'AAPL':1.0}
    start='2019-01-01'; end='2020-6-30'
    liquidity_type='roll_spread'
    liq1=plot_liquidity_monthly(portfolio,start,end,liquidity_type)       
    liquidity_type='amihud_illiquidity'
    liq2=plot_liquidity_monthly(portfolio,start,end,liquidity_type) 
    liquidity_type='ps_liquidity'
    liq3=plot_liquidity_monthly(portfolio,start,end,liquidity_type) 

    pf1={'Market':('US','^GSPC'),'VIPS':0.1,'PDD':0.2,'JD':0.3,'BABA':0.4}
    al1=plot_liquidity_monthly(pf1,start,end,'roll_spread') 

    pf2={'Market':('US','^GSPC'),'^GSPC':1.0}
    al2=plot_liquidity_monthly(pf2,start,end,'roll_spread')
    al3=plot_liquidity_monthly(pf2,start,end,'amihud_illiquidity')
    al4=plot_liquidity_monthly(pf2,start,end,'ps_liquidity')

#==============================================================================
def plot_liquidity_annual(portfolio,start,end,liquidity_type):
    """
    功能：将流动性指标逐年绘图对比
    输入：投资组合，开始/结束日期，流动性指标类别
    输出：流动性指标的逐年数据框
    显示：按年绘图投资组合的流动性指标
    """
    #仅为测试用
    #portfolio={'Market':('US','^GSPC'),'TSLA':1.0}
    #start='2009-07-01'
    #end='2019-06-30'
    #liquidity_type='roll_spread'    

    #检查日期和期间的合理性
    flag,start2,end2=check_period(start,end)
    if not flag:
        print("#Error(plot_liquidity_annual): invalid period for,", start, end)
        return None

    #检查：支持的liquidity_type
    liquidity_list=['roll_spread','amihud_illiquidity','ps_liquidity']
    if liquidity_type not in liquidity_list:
        print("#Error(plot_liquidity_annual): not supported liquidity type")
        print("Supported liquidity type:",liquidity_list)              
        return None         

    #抓取投资组合信息
    print("\n... Searching for portfolio information ...")
    df=get_portfolio_prices(portfolio,start,end)

    #拆分start/end之间的各个年份和月份
    mdlist=calc_yearly_date_range(start,end)
    if len(mdlist) == 0:
        print("#Error(plot_liquidity_annual): start/end dates inappropriate")
        return None          

    #用于保存指标
    print("... Calculating annual",liquidity_type,"...")
    rarfunc='calc_'+liquidity_type
    import pandas as pd
    rars=pd.DataFrame(columns=('YR','liquidity'))
    zeroline=False   
    for i in range(0,len(mdlist)):
        startyr=mdlist[i][0]
        YR=startyr.strftime("%Y")
        #print(YR,end=' ')
        endyr=mdlist[i][1]
        pfdf=sample_selection(df,startyr,endyr)
        rar=eval(rarfunc)(pfdf)        
        if rar is not None:     
            if rar < 0: zeroline=True
            row=pd.Series({'YR':YR,'liquidity':rar})
            rars=rars.append(row,ignore_index=True)         
    rars.set_index('YR',inplace=True)    
    
    #绘图
    colname='liquidity'
    collabel=ectranslate(liquidity_type)
    ylabeltxt=ectranslate(liquidity_type)
    titletxt="证券流动性风险的年度指标"

    _,_,tickerlist,sharelist=decompose_portfolio(portfolio)
    if len(tickerlist)==1:
        product=str(tickerlist)
    else:
        product=str(tickerlist)+' in '+str(sharelist)
    
    import datetime as dt; today=dt.date.today()
    footnote="证券="+product+"\n数据来源：雅虎财经, "+str(today)    
    datatag=False
    power=4
    plot_line(rars,colname,collabel,ylabeltxt,titletxt,footnote,datatag, \
              power,zeroline)

    return rars


if __name__=='__main__':
    portfolio={'Market':('US','^GSPC'),'AAPL':1.0}
    start='2010-01-01'
    end='2019-12-31'
    liquidity_type='roll_spread'
    al1=plot_liquidity_annual(portfolio,start,end,liquidity_type)  
    
    portfolio={'Market':('US','^GSPC'),'VIPS':0.1,'PDD':0.2,'JD':0.3,'BABA':0.4}
    start='2013-01-01'
    end='2019-06-30'    
    al2=plot_liquidity_annual(portfolio,start,end,liquidity_type) 


#==============================================================================
def draw_liquidity(liqs):
    """
    功能：绘制滚动窗口曲线
    输入：滚动数据df，内含投资组合Portfolio和指数名称Type
    """
    
    plt.figure(figsize=(8,5))

    portfolio=liqs['Portfolio'][0]
    colname='Ratio'
    collabel=ectranslate(liqs['Type'][0])
    ylabeltxt=ectranslate(liqs['Type'][0])
    titletxt="证券流动性风险的滚动趋势"
    
    _,_,tickerlist,sharelist=decompose_portfolio(portfolio)
    if len(tickerlist)==1:
        product=str(tickerlist)
    else:
        product=str(tickerlist)+' in '+str(sharelist)
    
    import datetime as dt; today=dt.date.today()
    footnote="证券="+product+"\n数据来源：雅虎财经, "+str(today)    
    datatag=False
    power=4
    
    zeroline=False
    neg_liqs=liqs[liqs['Ratio'] < 0]
    if len(neg_liqs) > 0: zeroline=True
    
    plot_line(liqs,colname,collabel,ylabeltxt,titletxt,footnote,datatag, \
              power,zeroline)

    return

#==============================================================================
def liquidity_rolling(portfolio,start,end,liquidity_type, \
                      window=30,graph=True):
    """
    功能：滚动计算一个投资组合的流动性风险
    投资组合的结构：{'Market':('US','^GSPC'),'AAPL':0.5,'MSFT':0.3,'IBM':0.2}
    输入：投资组合，开始日期，结束日期，指数名称，滚动窗口宽度(天数)
    输出：流动性风险
    注意：因需要滚动计算，开始和结束日期之间需要拉开距离
    """

    #检查日期和期间的合理性
    flag,start2,end2=check_period(start,end)
    if not flag:
        print("#Error(liquidity_rolling): invalid period for,", start, end)
        return None

    #检查：支持的liquidity_type
    liquidity_list=['roll_spread','amihud_illiquidity','ps_liquidity']
    if liquidity_type not in liquidity_list:
        print("#Error(liquidity_rolling): not supported liquidity type")
        print("Supported liquidity type:",liquidity_list)              
        return None         

    #抓取投资组合信息
    print("\n... Searching information for portfolio:",portfolio)
    reg=get_portfolio_prices(portfolio,start,end)
    if reg is None:
        print("#Error(liquidity_rolling): failed to get portfolio info")              
        return None          
    
    #滚动计算
    print("... Calculating its rolling ratios:",liquidity_type)
    import pandas as pd; import numpy as np
    datelist=reg.index.to_list()
    calc_func='calc_'+liquidity_type
    
    liqs=pd.DataFrame(columns=('Portfolio','Date','Ratio','Type')) 
    for i in np.arange(0,len(reg)):
        i1=i+window
        if i1 >= len(reg): break
        
        #构造滚动窗口
        windf=reg[reg.index >= datelist[i]]
        windf=windf[windf.index < datelist[i1]]
        #print(i,datelist[i],i1,datelist[i1],len(windf))
        
        #使用滚动窗口计算
        liq=eval(calc_func)(windf)
        if liq is None: continue
        
        #记录计算结果
        row=pd.Series({'Portfolio':portfolio,'Date':datelist[i1-1],'Ratio':liq,'Type':liquidity_type})
        liqs=liqs.append(row,ignore_index=True)        
    
    liqs.set_index(['Date'],inplace=True) 
    
    #第6步：绘图
    if graph == True:
        draw_liquidity(liqs)
    
    return liqs


if __name__=='__main__':
    portfolio={'Market':('US','^GSPC'),'AAPL':0.5,'MSFT':0.3,'IBM':0.2}
    start='2019-1-1'
    end='2019-6-30'
    window=30
    graph=True    
    liquidity_type='roll_spread'
    liqs=liquidity_rolling(portfolio,start,end,liquidity_type)


#==============================================================================
#==============================================================================
def compare_liquidity_rolling(portfolio1,portfolio2,start,end,liquidity_type,window=30):
    """
    功能：比较两个投资组合portfolio1和portfolio2在start至end期间的流动性指标liquidity_type
    """
    #投资组合1
    liqs1=liquidity_rolling(portfolio1,start,end,liquidity_type, \
                      window=window,graph=False)
    if liqs1 is None:
        print("#Error(compare_liquidity_rolling): portfolio info inaccesible for",portfolio1)
        return 
    
    colname1='Ratio'
    label1=ectranslate(liquidity_type)
    datatag1=False
    
    zeroline=False
    neg_liqs1=liqs1[liqs1['Ratio'] < 0]
    if len(neg_liqs1) > 0: zeroline=True
   
    #投资组合2
    liqs2=liquidity_rolling(portfolio2,start,end,liquidity_type, \
                      window=window,graph=False)
    if liqs2 is None:
        print("#Error(compare_liquidity_rolling): portfolio info inaccesible for",portfolio2)
        return     
    
    colname2='Ratio'
    label2=ectranslate(liquidity_type)
    datatag2=False
    
    zeroline=False
    neg_liqs2=liqs2[liqs2['Ratio'] < 0]
    if len(neg_liqs2) > 0: zeroline=True

    #绘图    
    ylabeltxt=ectranslate(liquidity_type)
    titletxt="证券流动性风险的滚动趋势比较"
    
    _,_,tickerlist1,sharelist1=decompose_portfolio(portfolio1)
    if len(tickerlist1)==1:
        product1=str(tickerlist1)
    else:
        product1=str(tickerlist1)+'，持仓'+str(sharelist1)    
    _,_,tickerlist2,sharelist2=decompose_portfolio(portfolio2)
    if len(tickerlist2)==1:
        product2=str(tickerlist2)
    else:
        product2=str(tickerlist2)+'，持仓'+str(sharelist2)     
    
    import datetime as dt; today=dt.date.today()
    footnote="证券1="+product1+"\n证券2="+product2+ \
            "\n数据来源：雅虎财经, "+str(today)
    power=0 #不绘制趋势线

    plot_line2_coaxial(liqs1,"证券1",colname1,label1, \
                       liqs2,"证券2",colname2,label2, \
                ylabeltxt,titletxt,footnote,power,datatag1,datatag2,zeroline)
    return
    
    

if __name__ =="__main__":
    portfolio1={'Market':('US','^GSPC'),'PDD':1.0}
    portfolio2={'Market':('US','^GSPC'),'BILI':1.0}
    start='2020-1-01'
    end='2020-6-30'
    window=21
    graph=False
    liquidity_type='roll_spread'   
    compare_liquidity_rolling(portfolio1,portfolio2,start,end,liquidity_type)

    liquidity_type='amihud_illiquidity'   
    compare_liquidity_rolling(portfolio1,portfolio2,start,end,liquidity_type)    

    liquidity_type='ps_liquidity'   
    compare_liquidity_rolling(portfolio1,portfolio2,start,end,liquidity_type)    
#==============================================================================
#==============================================================================





# -*- coding: utf-8 -*-

#==============================================================================
#关闭所有警告
import warnings; warnings.filterwarnings('ignore')
from siat.common import *
from siat.grafix import *
#==============================================================================
if __name__=='__main__':
    symbol='黄金期权'

def option_comm_dict_china(symbol):
    """
    获取中国商品期权大类合约
    显示symbol类期权的可用合约列表
    """
    import akshare as ak
    import datetime
    today = datetime.date.today()
    
    try:
        optiondict=ak.option_sina_commodity_dict(symbol=symbol)
        print("\n中国"+symbol+"的可用合约：")
        contractlist=optiondict[symbol]
        contractlist.sort(reverse=False)
        #print(contractlist)
        printlist(contractlist,numperline=8,beforehand=' '*4,separator=' ')
        print('*** 注：合约代码后四位数字为合约到期日YYMM')
        print('    每种合约还将分为看涨(C)/看跌(P)两个方向和不同的行权价')
        print('    数据来源：新浪财经,',today)
        return optiondict
    except:
        print("  #Error(option_com_china): failed to get dict info for",symbol)
        print("  Solution: upgrade siat and akshare plug-in, then try again")
        print("  If problem remains, report to the author via wechat or email.")
        return None    

#==============================================================================
if __name__=='__main__':
    underlying='黄金'
    contract='au2204'
def option_comm_china(underlying='',contract=''):
    """
    获取中国商品期权大类
    若underlying=''或错误且contract=''时显示中国商品期权大类
    若underlying查到且contract=''或未查到时显示该类期权的合约列表
    若underlying查到且contract查到时显示该类期权合约的价格与持仓量分析
    """
    import akshare as ak
    import datetime
    today = datetime.date.today()
    
    underlyinglist=['豆粕','玉米','铁矿石','棉花','白糖','PTA','甲醇','橡胶','沪铜','黄金','菜籽粕','液化石油气','动力煤']
    if not (underlying in underlyinglist):
        print("\n中国商品期权的常见品种：")
        #print(underlyinglist)
        printlist(underlyinglist,numperline=6,beforehand=' '*4,separator=' ')
        print('*** 数据来源：新浪财经,',today)
        return underlyinglist
    
    symbol=underlying+'期权'
    if (underlying in underlyinglist) and (contract==''):
        optiondict=option_comm_dict_china(symbol)
        return optiondict
        
    try:
        df2=ak.option_sina_commodity_contract_list(symbol=symbol,contract=contract)
    except:
        print("  #Error(option_com_china): contract",contract,'not found in',symbol)
        optiondict=option_comm_dict_china(symbol)
        return optiondict 
    
    df2cols=['买量C','买价C','最新价C','卖价C','卖量C','持仓量C','涨跌C','行权价','看涨期权合约','买量P','买价P','最新价P','卖价P','卖量P','持仓量P','涨跌P','看跌期权合约']
    df2.columns=df2cols

    df2['最新价C']=df2['最新价C'].astype('float')
    df2['持仓量C']=df2['持仓量C'].astype('float')
    df2['最新价P']=df2['最新价P'].astype('float')
    df2['持仓量P']=df2['持仓量P'].astype('float')
    df2['行权价']=df2['行权价'].astype('float')
    df2.set_index('行权价',inplace=True)
    
    df2c=df2['看涨期权合约']
    df2c.dropna(inplace=True)
    df2clist=list(df2c)
    print("  \n中国"+symbol+contract+"的看涨期权合约：")
    printlist(df2clist,numperline=6,beforehand=' '*4,separator=' ')
    df2p=df2['看跌期权合约']
    df2p.dropna(inplace=True)
    df2plist=list(df2p)
    print("  \n中国"+symbol+contract+"的看跌期权合约：")  
    printlist(df2plist,numperline=6,beforehand=' '*4,separator=' ')

    footnote="行权价-->\n\n"+"数据来源：新浪财经，"+str(today)

    print("\nRendering graphics for the relationships btw contract and strike prices...")
    titletxt="当前期权价格与行权价的关系："+symbol+contract
    plot_line2(df2,"看涨期权",'最新价C','价格', \
                 df2,"看跌期权",'最新价P','价格', \
                 '价格',titletxt,footnote,power=0,twinx=False)
    """
    print("Rendering graphics for the relationships btw call price and open interest ...")
    titletxt="当前期权价格与持仓量的关系："+symbol+contract+'的'+"看涨期权"
    plot_line2(df2,"看涨期权",'最新价C','价格', \
                 df2,"看涨期权",'持仓量C','持仓量', \
                 '',titletxt,footnote,power=0,twinx=True)

    print("Rendering graphics for the relationships btw put price and open interest ...")
    titletxt="当前期权价格与持仓量的关系："+symbol+contract+'的'+"看跌期权"
    plot_line2(df2,"看跌期权",'最新价P','价格', \
                 df2,"看跌期权",'持仓量P','持仓量', \
                 '',titletxt,footnote,power=0,twinx=True)

    print("Rendering graphics for the relationships btw open interests ...")
    titletxt="当前期权方向与持仓量的关系："+symbol+contract
    plot_line2(df2,"看涨期权",'持仓量C','持仓量', \
                 df2,"看跌期权",'持仓量P','持仓量', \
                 '',titletxt,footnote,power=0,twinx=False)
    """
    return df2
#==============================================================================
#==============================================================================
if __name__=='__main__':
    contract='au2112C328'
    contract=['au2112C328','au2112P328']
    contract=[]
    power=0
    twinx=False
    start='2021-8-1'
    end='2021-8-31'

def option_comm_trend_china(contract,start='',end='',power=0,twinx=False):
    """
    绘制期权合约价格历史价格趋势图
    若contract为一个合约，则绘制单折线图
    若contract为多于一个合约，则绘制前两个合约的双单折线图
    """
    
    contract1=contract2=''
    if isinstance(contract,str):
        contract1=contract
        contract2=''
    
    if isinstance(contract,list):
        if len(contract)==1:
            contract1=contract
            contract2=''    
        elif len(contract)>=2:
            contract1=contract[0]
            contract2=contract[1]    
            
    if contract1=='':
        print("  #Error(option_comm_trend_china): unknown option contract",contract)
        return None

    import pandas as pd
    start1=end1=''
    if not (start==''):
        try:
            start1=pd.to_datetime(start)
        except:
            print("  #Error(option_comm_trend_china): invalid date",start)
    if not (end==''):
        try:
            end1=pd.to_datetime(end)
        except:
            print("  #Error(option_comm_trend_china): invalid date",end)
    
    import akshare as ak
    import datetime
    today = datetime.date.today()
    footnote="数据来源：新浪财经，"+str(today)
    
    #绘制单折线
    if contract2=='':
        contract1,dict1=option_contract_decode_china(contract1)
        try:
            df3=ak.option_sina_commodity_hist(contract=contract1)    
        except:
            print("  #Error(option_comm_trend_china): contract",contract1,"not found")
            return None
        
        if len(df3)==0:
            print("  #Warning(option_comm_trend_china): no record found for contract",contract1)
            return None            
        
        df3['date2']=pd.to_datetime(df3['date'])
        df3.set_index('date2',inplace=True)
        df3['close']=df3['close'].astype('float')
        
        if not (start1==''):
            df3=df3.drop(df3[df3.index < start1].index)
        if not (end1==''):
            df3=df3.drop(df3[df3.index > end1].index)        

        print("  Rendering graphics for option contract price trend...")
        titletxt="期权合约价格的运动趋势："+contract1
        footnote=contract1+'：'+dict1['标的物']+dict1['期权方向']+'，'+dict1['到期日']+'到期'+'，行权价'+dict1['行权价']+'\n'+footnote
        plot_line(df3,'close','收盘价','价格',titletxt,footnote,power=power)
        return df3
    
    #绘制双折线
    contract1ok=contract2ok=True
    if not (contract2==''):
        contract1,dict1=option_contract_decode_china(contract1)
        try:
            df31=ak.option_sina_commodity_hist(contract=contract1)   
            df31['date2']=pd.to_datetime(df31['date'])
            df31.set_index('date2',inplace=True)
            df31['close']=df31['close'].astype('float')
        except:
            contract1ok=False
        if contract1ok:
            if not (start1==''):
                df31=df31.drop(df31[df31.index < start1].index)
            if not (end1==''):
                df31=df31.drop(df31[df31.index > end1].index)            

        if len(df31)==0:
            #print("  #Warning(option_comm_trend_china): no record found for contract",contract1)
            contract1ok=False
        
        contract2,dict2=option_contract_decode_china(contract2)
        try:
            df32=ak.option_sina_commodity_hist(contract=contract2) 
            df32['date2']=pd.to_datetime(df32['date'])
            df32.set_index('date2',inplace=True)
            df32['close']=df32['close'].astype('float')
        except:
            contract2ok=False
        if contract2ok:
            if not (start1==''):
                df32=df32.drop(df32[df32.index < start1].index)
            if not (end1==''):
                df32=df32.drop(df32[df32.index > end1].index)   

        if len(df32)==0:
            #print("  #Warning(option_comm_trend_china): no record found for contract",contract2)
            contract2ok=False
        
        if contract1ok and contract2ok:
            print("  Rendering graphics for comparing two option contract price trends...")
            titletxt="期权价格的运动趋势对比："+contract1+'与'+contract2
            footnote=contract2+'：'+dict2['标的物']+dict2['期权方向']+'，'+dict2['到期日']+'到期'+'，行权价'+dict2['行权价']+'\n'+footnote
            footnote=contract1+'：'+dict1['标的物']+dict1['期权方向']+'，'+dict1['到期日']+'到期'+'，行权价'+dict1['行权价']+'\n'+footnote
            
            plot_line2(df31,contract1,'close','收盘价', \
                   df32,contract2,'close','收盘价', \
                       '价格',titletxt,footnote,twinx=twinx)
            return df31,df32
        elif contract1ok:
            print("  Rendering graphics for option contract price trend...")
            titletxt="期权合约价格的运动趋势："+contract1
            footnote=contract1+'：'+dict1['标的物']+dict1['期权方向']+'，'+dict1['到期日']+'到期'+'，行权价'+dict1['行权价']+'\n'+footnote
            
            plot_line(df31,'close','收盘价','价格',titletxt,footnote,power=power)
            return df31            
        elif contract2ok:
            print("  Rendering graphics for option contract price trend...")
            titletxt="期权合约价格的运动趋势："+contract2
            footnote=contract2+'：'+dict2['标的物']+dict2['期权方向']+'，'+dict2['到期日']+'到期'+'，行权价'+dict2['行权价']+'\n'+footnote
            
            plot_line(df32,'close','收盘价','价格',titletxt,footnote,power=power)
            return df32  
        else:
            print("  #Warning(option_comm_trend_china): no record found for contracts",contract1,'and',contract2)
            return None
        
#==============================================================================
if __name__=='__main__':
    contract='xu2112'
    contract='au2112C328'
    
def option_contract_decode_china(contract):
    """
    例：
    contract='c2111'或'cf2111'
    contract='c2111C235'或'cf2111P235'
    """
    
    prelist=['m','c','i','cf','sr','ta','ma','ru','cu','au','rm','pg','zc']
    ualist=['豆粕','玉米','铁矿石','棉花','白糖','PTA','甲醇','橡胶','沪铜','黄金','菜籽粕','液化石油气','动力煤']
    
    import string
    ucletters=list(string.ascii_uppercase)    
    lcletters=list(string.ascii_lowercase)
    
    pos=0
    contract1=contract.lower()
    for c in contract1:
        if c in lcletters:
            pos=pos+1
        else:
            break
    prefix=contract1[:pos]
    yymm=contract1[pos:pos+4]
    maturity='20'+yymm[:2]+'-'+yymm[2:] #到期年月
    
    direction=''
    strike=''
    if len(contract1)>6:
        direction=contract1[pos+4:pos+5]
        direction=direction.upper()     #期权方向
        contract2=contract1[:pos+4]+direction+contract1[pos+5:]
        strike=contract1[pos+5:]        #行权价 

        if direction=='C':
            otype="看涨期权"
        elif direction=='P':
            otype="看跌期权"
        else:
            otype="未知"
    else:
        contract2=contract1
    
    try:
        pos1=prelist.index(prefix)
        ua=ualist[pos1]                 #期权标的物类别
    except:
        print("  #Error(option_contract_decode_china): contract",contract,"not found")
        return None,None
    
    contract_notes={}
    contract_notes['合约']=contract2
    contract_notes['标的物']=ua
    contract_notes['到期日']=maturity
    if not (direction==''):
        contract_notes['期权方向']=otype
    if not (strike==''):
        contract_notes['行权价']=strike
    
    
    return contract2,contract_notes




# -*- coding: utf-8 -*-

import os; os.chdir("S:\siat")
from siat.assets_liquidity import *

portfolio={'Market':('China','000001.SS'),'600011.SS':1}
start='2020-1-1'
end='2020-6-30'
liquidity_type='roll_spread'
l=liquidity_rolling(portfolio,start,end,liquidity_type,30)

ticker=['600011.SS']
start='2020-1-1'
end='2020-6-30'
pak=get_prices_ak(ticker,start,end)

pyf=get_price_yf(ticker,start,end)

pyh=p=get_prices_yahoo(ticker,start,end)

p=get_prices(ticker,start,end)

tickerlist=['600011.SS']
sharelist=[1]
p1=get_price_portfolio(tickerlist,sharelist,start,end)


if __name__=='__main__':
    tickerlist=['INTC','MSFT']
    sharelist=[0.6,0.4]
    fromdate='2020-11-1'
    todate='2021-1-31'

p2=get_prices_portfolio(tickerlist,sharelist,fromdate,todate)





# -*- coding: utf-8 -*-


import os; os.chdir("S:/siat")
from siat.markowitz import *

Market={'Market':('US','^GSPC')}
Stocks={'AAPL':.1,'MSFT':.13,'XOM':.09,'JNJ':.09,'JPM':.09,'AMZN':.15,'GE':.08,'FB':.13,'T':.14}
portfolio=dict(Market,**Stocks)

pf_info=portfolio_cumret(portfolio,'2019-12-31')
portfolio_covar(pf_info)
portfolio_corr(pf_info)
portfolio_expectation(pf_info)

es_info=portfolio_es(pf_info,simulation=50000)

df=portfolio_MSR_GMV(es_info)

stocks=['IBM','WMT','AAPL','C','MSFT']
ef=portfolio_ef(stocks,'2019-1-1','2020-8-1')





_,_,tickerlist,sharelist=decompose_portfolio(portfolio)

today='2020-12-31'
pastyears=1

pf_info=portfolio_cumret(portfolio,'2020-12-31')

portfolio_covar(pf_info)

portfolio_corr(pf_info)


















#定义投资组合
Market={'Market':('US','^GSPC')}
Stocks={'BABA':.4,'JD':.3,'PDD':.2,'VIPS':.1}
portfolio=dict(Market,**Stocks)

#搜寻该投资组合中所有成分股的价格信息，默认观察期为一年，pastyears=1
pf_info=portfolio_cumret(portfolio,'2020-11-30',pastyears=1)

#生成了投资组合的可行集
es_info=portfolio_es(pf_info,simulation=50000)
es_info10=portfolio_es(pf_info,simulation=100000)

#寻找投资组合的MSR优化策略点和GMV优化策略点
psr=portfolio_MSR_GMV(es_info)



# -*- coding: utf-8 -*-
"""
本模块功能：数字货币及其MSA交易策略
所属工具包：证券投资分析工具SIAT 
SIAT：Security Investment Analysis Tool
创建日期：2018年9月19日
最新修订日期：2020年2月3日
作者：王德宏 (WANG Dehong, Peter)
作者单位：北京外国语大学国际商学院
版权所有：王德宏
用途限制：仅限研究与教学使用，不可商用！商用需要额外授权。
特别声明：作者不对使用本工具进行证券投资导致的任何损益负责！
"""

#关闭所有警告
import warnings; warnings.filterwarnings('ignore')
#引用Python插件 
from datetime import datetime
import numpy as np; import pandas as pd
#from scipy import stats
import json
from bs4 import BeautifulSoup; import requests
#==============================================================================
import matplotlib.pyplot as plt
#处理绘图汉字乱码问题
import sys; czxt=sys.platform
if czxt in ['win32','win64']:
    #设置绘图时的汉字显示
    #plt.rcParams['font.sans-serif'] = ['FangSong']  # 设置默认字体
    plt.rcParams['font.sans-serif'] = ['SimHei']  # 设置默认字体
if czxt in ['darwin']:
    plt.rcParams['font.sans-serif']=['Arial Unicode MS']
    #plt.rcParams['font.sans-serif']= ['Heiti TC']
# 解决保存图像时'-'显示为方块的问题
plt.rcParams['axes.unicode_minus'] = False 
#==============================================================================

#定义爬虫函数：抓取交易所信息
def fetchCrypto_Exchange(fsym,tsym,top=5):
    urlprefix="https://min-api.cryptocompare.com/data/top/exchanges/full?fsym="
    url      = urlprefix + fsym + "&tsym=" + tsym       
    response = requests.get(url)
    soup     = BeautifulSoup(response.content, "html.parser")
    dic      = json.loads(soup.prettify())

    #筛选可交易所选产品的交易所，并显示列表
    market = []
    d = dic['Data']['Exchanges']
    for i in range(len(d)):
        market.append(d[i]['MARKET'])

    #基于过去24小时内交易量进行交易所排名
    vol = []
    d = dic['Data']['Exchanges']
    for i in range(len(d)):
        volamt=round(float(d[i]['VOLUME24HOUR']),2)
        if volamt > 0:
            vol.append([d[i]['MARKET'], volamt])
 
    #基于子列表中的第二项对子列表排序
    vol = sorted(vol, key=lambda x: -x[1])
    
    #基于过去24小时交易量显示所选产品活跃的5个交易所名单
    print("\n全球最活跃的五家数字货币交易所：",fsym+"/"+tsym,"\n       (24小时交易金额，"+tsym+"百万)")
    for e in vol:
        print("%10s%15.2f" % (e[0], e[1]))
    import datetime as dt; nowtime=dt.datetime.now().strftime('%Y-%m-%d %H:%M')
    print("\n数据来源：CryptoCompare,",nowtime)    
    #活跃交易所Top 5，markets
    markets = [e[0] for e in vol][0:top]
    """
    print("\n",markets,"\n")
    """
    return markets

if __name__=='__main__':
    fsym="ETH"
    tsym="USD"
    top=10
    markets=fetchCrypto_Exchange("ETH","USD")

#==============================================================================

#定义爬虫函数：从指定交易所抓取所选产品价格
#限制：考虑到网络传输量，一次抓取最多2000条交易（2000/365=5.5年最多）
def fetchCrypto_Price_byExchange(fsym, tsym, exchange):
    #数据源: https://www.cryptocompare.com/api/
    cols = ['date', 'timestamp', 'open', 'high', 'low', 'close']
    lst = ['time', 'open', 'high', 'low', 'close']
 
    timestamp_today = datetime.today().timestamp()
    curr_timestamp = timestamp_today
 
    for j in range(2):
        df = pd.DataFrame(columns=cols)
        urlprefix="https://min-api.cryptocompare.com/data/histoday?fsym="
        url = urlprefix + fsym + \
              "&tsym=" + tsym + "&toTs=" + str(int(curr_timestamp)) + \
              "&limit=2000" + "&e=" + exchange
        response = requests.get(url)
        soup = BeautifulSoup(response.content, "html.parser")
        dic = json.loads(soup.prettify())        
        if len(dic['Data']) < 1:
            return None    #爬虫失败，返回空
        
        for i in range(1, 2001):
            tmp = []
            for e in enumerate(lst):
                x = e[0]; y = dic['Data'][i][e[1]]
                if(x == 0):
                    #将timestamp转换为日期
                    td = datetime.fromtimestamp(int(y)).strftime('%Y-%m-%d')
                    tmp.append(td)
                tmp.append(y)
            if(np.sum(tmp[-4::]) > 0):
                df.loc[len(df)] = np.array(tmp)
                
        df.index = pd.to_datetime(df.date)
        df.drop('date', axis=1, inplace=True)
        curr_timestamp = int(df.iloc[0][0])
 
        if(j == 0):
            df0 = df.copy()
        else:
            data = pd.concat([df, df0], axis=0)
 
    return data.astype(np.float64)

if __name__=='__main__':
    data=fetchCrypto_Price_byExchange("ETH", "USD", "Coinsbit")


#==============================================================================
def Crypto_Price_Trend(fsym, tsym, exchange,fromdate,todate,power=4):
    """
    功能：绘制价格趋势线
    fsym: 数字货币产品
    tsym: 交易币种
    exchange: 交易所
    
    """
    print("... Searching for information, please wait ...")
    
    #检查日期期间是否有效
    import siat.common as cmn
    result, start, end=cmn.check_period(fromdate, todate)
    if not result:
        print("#Error(Crypto_Price_Trend): invalid date period from",fromdate,'to',todate)
        return None

    #获得价格
    p=fetchCrypto_Price_byExchange(fsym, tsym, exchange)
    p1=p[p.index >= start]
    p2=p1[p1.index <= end]

    import siat.grafix as gfx        
    colname='close'
    collabel='收盘价'
    ylabeltxt='价格('+tsym+')'
    titletxt="数字货币价格趋势："+fsym+'/'+tsym+'，'+exchange
    import datetime as dt; nowtime=dt.datetime.now().strftime('%Y-%m-%d %H:%M')
    footnote="数据来源：CryptoCompare,"+str(nowtime)        
    
    import siat.grafix as g
    g.plot_line(p2,colname,collabel,ylabeltxt,titletxt,footnote,power=power)
    
    return p2

if __name__=='__main__':
    fsym="ETH"
    tsym="USD"
    exchange='Coinbase'
    fromdate='2020-7-1'
    todate='2020-12-31'
    power=3
    price=Crypto_Price_Trend(fsym, tsym, exchange,fromdate,todate,power)
#==============================================================================
def compCrypto_Price(product1,product2,days=30):
    """
    功能：比较两种数字货币的价格趋势，最近days天的情形
    输入：
    product1/product2：产品列表，格式：[数字货币价格数据框,产品,币种,市场]
    输出：
    绘制折线图
    无返回数据
    """
    
    #绘制折线图  
    label1=product1[1]+'/'+product1[2]+' @'+product1[3]
    p1=product1[0].tail(days)
    plt.plot(p1['close'],label=label1,lw=3)
    
    label2=product2[1]+'/'+product2[2]+' @'+product2[3]
    p2=product2[0].tail(days)
    plt.plot(p2['close'],label=label2,lw=3,ls=':')    

    #图示标题
    titletxt="数字货币产品：跨市场的价格套利机会"
    plt.title(titletxt,fontweight='bold')
    plt.ylabel("收盘价")
    plt.xticks(rotation=45)
    plt.legend(loc='best')
    plt.show()
    
    return    
    
if __name__=='__main__':
    prices1=fetchCrypto_Price_byExchange("ETH", "USD", "Coinsbit")
    product1=[prices1,"ETH", "USD", "Coinsbit"]
    prices2=fetchCrypto_Price_byExchange("ETH", "USD", "Coinbase")
    product2=[prices2,"ETH", "USD", "Coinbase"]
    compCrypto_Price(product1,product2)    


#==============================================================================
def compCrypto_Return(product1,product2,days=30):
    """
    功能：比较两种数字货币的收益率趋势，最近days天的情形
    输入：
    product1/product2：产品列表，格式：[数字货币价格数据框,产品,币种,市场]
    输出：
    绘制折线图
    无返回数据
    """
    import pandas as pd
    
    #绘制折线图      
    r1=product1[0]['close'].pct_change()
    r1df=pd.DataFrame(r1)
    r1df.columns=['ret']
    r1df['ret%']=round(r1df['ret']*100.0,2)    
    r1t=r1df.tail(days)
    label1=product1[1]+'/'+product1[2]+' @'+product1[3]
    plt.plot(r1t['ret%'],label=label1,lw=3)
    
    r2=product2[0]['close'].pct_change()
    r2df=pd.DataFrame(r2)
    r2df.columns=['ret']
    r2df['ret%']=round(r2df['ret']*100.0,2)    
    r2t=r2df.tail(days)    
    label2=product2[1]+'/'+product2[2]+' @'+product2[3]    
    plt.plot(r2t['ret%'],label=label2,lw=3,ls=':')  
    
    plt.axhline(y=0.0,color='green',linestyle='--')

    #图示标题
    titletxt="数字货币产品：跨市场的收益率套利机会"
    plt.title(titletxt,fontweight='bold')
    plt.ylabel("资本利得%")
    plt.xticks(rotation=45)
    plt.legend(loc='best')
    plt.show()
    
    return    
    
if __name__=='__main__':
    prices1=fetchCrypto_Price_byExchange("ETH", "USD", "Coinsbit")
    product1=[prices1,"ETH", "USD", "Coinsbit"]
    prices2=fetchCrypto_Price_byExchange("ETH", "USD", "Coinbase")
    product2=[prices2,"ETH", "USD", "Coinbase"]
    compCrypto_Return(product1,product2)    


#==============================================================================
#定义爬虫函数：抓取列表中每个交易所的价格数据，合并数据到cp中
#注意：在指定的时间段，并非每个交易所都有ETH交易，若无则以nan表示
def fetchCrypto_Price_byExchList(fsym,tsym,markets,begdate,enddate):
    cp=pd.DataFrame([]); print("%s/%s" % (fsym, tsym))
    for market in markets:
        print("%12s... " % market, end="")
        df = fetchCrypto_Price_byExchange(fsym, tsym, market)
        if df is None:
            print("download failed")
            continue   #爬虫失败，跳出本次循环，继续下一轮循环

        ts = df[(df.index >= begdate) & (df.index <= enddate)]["close"]
        ts.name = market
        if ('cp' in globals()) or ('cp' in locals()): #已经存在cp
            tsdf=pd.DataFrame(ts)
            cp = pd.merge(cp,tsdf,how='outer',left_index=True,right_index=True)
        else:    #第一次生成cp
            cp = pd.DataFrame(ts)
        print("downloaded")
    return cp
   
#==============================================================================

#定义函数：计算2个市场组合之间所选产品的价差估计：均值，标准差
def calcSpread_in2Markets(cp):
    dist = []   #存放每两个交易所组合之间价差的均值和标准差
    for i in range(cp.shape[1]):
        for j in range(i):
            if(i != j):
                x = np.array(cp.iloc[:,i], dtype=np.float32)
                y = np.array(cp.iloc[:,j], dtype=np.float32)
                diff = np.abs(x-y)
                avg = np.mean(diff)
                std = np.std(diff, ddof=1)
                dist.append([cp.columns[i], cp.columns[j], avg, std])
 
    dist=pd.DataFrame(dist)
    dist.columns=['Market1','Market2','avg','std']
    dist=dist.dropna(axis=0,how='any')    #删除带有nan的行
    dist1=dist.sort_values(by=['avg','std'],ascending=(False,False))
    dist1=dist1.reset_index(drop = True)  #重新索引
    dist2=dist1.sort_values(by=['std','avg'],ascending=(False,False))

    return dist1,dist2

#==============================================================================

#定义函数：显示2个市场组合之间所选产品的价差估计：均值，标准差
def printSpread_in2Markets(dist1,dist2):
    
    dist1.rename(columns={'Market1':'市场1','Market2':'市场2','avg':'价差均值','std':'价差标准差'},inplace=True)
    dist2.rename(columns={'Market1':'市场1','Market2':'市场2','avg':'价差均值','std':'价差标准差'},inplace=True)
    dist2b=dist2[['市场1','市场2','价差标准差','价差均值']]
    
    import pandas as pd
    pd.set_option('display.unicode.ambiguous_as_wide', True)
    pd.set_option('display.unicode.east_asian_width', True)
    pd.set_option('display.width', 180) # 设置打印宽度(**重要**)

    print("\n ===== 市场间价差大小：按均值降序排列 =====")
    print(dist1)
    print("\n===== 市场间价差风险：按标准差降序排列 =====")
    print(dist2b)    
    
    dist2b['收益-风险性价比']=round(dist2b['价差均值']/dist2b['价差标准差'],4)
    dist3=dist2b[['市场1','市场2','收益-风险性价比','价差标准差','价差均值']]
    dist3=dist3.sort_values(by=['收益-风险性价比'],ascending=(False))
    print("\n       ===== 市场间价差：按收益-风险性价比降序排列 =====")
    print(dist3)       
    
    return 
#==============================================================================

#定义函数：绘图所选市场组合中的价差分布
def evalSpread_in2Markets(fsym,tsym,market1,market2,begdate,enddate):
    #抓取两个市场的价格数据 
    df1 = fetchCrypto_Price_byExchange(fsym, tsym, market1)
    df2 = fetchCrypto_Price_byExchange(fsym, tsym, market2)
 
    #过滤时间段
    df1 = df1[(df1.index > begdate) & (df1.index <= enddate)]
    df2 = df2[(df2.index > begdate) & (df2.index <= enddate)]
 
    #检查：两个市场在所选时间段每日必须都有交易，即收盘价数据表必须形状相同
    #如果数据表形状不同，需要寻找双方共同具有交易数据的时间段，并重设时间段
    #print(df1.close.shape[0], df2.close.shape[0])
 
    #绘图：观察两个市场间的价差分布，再次检查是否存在套利机会：曲线不完全重叠
    plt.figure(figsize=(12,6))
    plt.plot(df1.close, '.-', label=market1)
    plt.plot(df2.close, '.-', label=market2)
    plt.legend(loc=2)
    plt.title(fsym, fontsize=12)
    plt.ylabel(tsym, fontsize=12)
    plt.grid()

    return df1,df2
    
#==============================================================================

#定义函数：回测MSA投资策略 
def backtestMSA_Strategy(investment,account1,account2,position,df1,df2):
    roi = []                #投资回报率记录
    money = []              #资金头寸记录
    ac1 = [account1]; ac2 = [account2]
    #各个交易所的盈亏记录；pnl表示盈亏Profit and Loss
    pnl_exch1 = []; pnl_exch2 = []
    trade = False           #标识当前是否处于交易状态
    n = df1.close.shape[0]  #所选产品的收盘价样本个数
    trade_pnl=[]            #整体投资的盈亏记录
 
    #交易回测
    for i in range(n):  #对于每个日交易价格样本循环
        p1 = float(df1.close.iloc[i]); p2 = float(df2.close.iloc[i])
        if(p1 > p2):    #[if-A]若交易所1中产品的价格高于交易所2，卖出1买入2
            asset1 = "SHORT"; asset2 = "LONG"
            if(trade == False):
                #若当前未处于交易状态，则开始一个新的交易
                open_p1 = p1; open_p2 = p2
                open_asset1 = asset1; open_asset2 = asset2
                trade = True        #标识：当前处于交易状态
                print("new traded opened:")
                new_trade = False   #标识：当前已处于交易状态，并非新交易
            elif(asset1 == open_asset1):
                new_trade = False
            elif(asset1 == open_asset2):
                new_trade = True
 
        elif(p2 > p1):  #[if-A]若交易所1中产品的价格低于交易所2，买入1卖出2
            asset1 = "LONG"; asset2 = "SHORT"
            if(trade == False):
                #若当前未处于交易状态，则开始一个新的交易
                open_p1 = p1; open_p2 = p2
                open_asset1 = asset1; open_asset2 = asset2
                trade = True
                print("new traded opened:")
                new_trade = False
            elif(asset1 == open_asset1):
                new_trade = False
            elif(asset1 == open_asset2):
                new_trade = True
 
        if(i == 0):         #[if-B]抓取交易样本的起点
            print(df1.close.iloc[i], df2.close.iloc[i], \
                  asset1, asset2, trade, "----first trade info")
        else:               #[if-B]当前并非交易样本的起点，处于交易中间                   
            if(new_trade):  #[if-B:if-B1]当前处于交易空档期
                #核算当前账户头寸:交易所1卖出
                if(open_asset1 == "SHORT"): #[if-B:if-B1:if-B11]若交易所1卖出
                    #计算交易盈亏
                    pnl_asset1 = open_p1/p1 - 1
                    pnl_asset2 = p2/open_p2 -1
                    pnl_exch1.append(pnl_asset1)
                    pnl_exch2.append(pnl_asset2)
                    print(open_p1, p1, open_p2, p2, open_asset1, \
                          open_asset2, pnl_asset1, pnl_asset2)
                    #更新账户余额
                    account1 = account1 + position*pnl_asset1
                    account2 = account2 + position*pnl_asset2
                    print("accounts [USD] = ", account1, account2)
                    if((account1 <=0) or (account2 <=0)):
                        print("--trading halted")   #账户余额不足，停止交易
                        break
                    #计算投资回报ROI
                    total = account1 + account2
                    roi.append(total/investment-1)
                    ac1.append(account1); ac2.append(account2)
                    money.append(total)
                    print("ROI = ", roi[-1]); print("trade closed\n")
                    trade = False
 
                    #开始一个新的交易
                    if(asset1 == "SHORT"):  #交易所1卖出交易
                        open_p1 = p1; open_p2 = p2
                        open_asset1 = asset1; open_asset2 = asset2
                    else:
                        open_p1 = p1; open_p2 = p2
                        open_asset1 = asset1; open_asset2 = asset2
                    trade = True
                    print("new trade opened:", asset1, asset2, \
                          open_p1, open_p2)
 
                #核算当前账户头寸:交易所1买入
                if(open_asset1 == "LONG"):  #[if-B:if-B1:if-B12]若交易所1买入
                    #计算交易盈亏
                    pnl_asset1 = p1/open_p1 -1
                    pnl_asset2 = open_p2/p2 - 1
                    pnl_exch1.append(pnl_asset1)
                    pnl_exch2.append(pnl_asset2)
                    print(open_p1, p1, open_p2, p2, open_asset1, \
                          open_asset2, pnl_asset1, pnl_asset2)
                    #更新账户余额
                    account1 = account1 + position*pnl_asset1
                    account2 = account2 + position*pnl_asset2
                    print("accounts [USD] = ", account1, account2)
                    if((account1 <=0) or (account2 <=0)):
                        print("--trading halted")
                        break
                    #计算投资回报ROI
                    total = account1 + account2
                    roi.append(total/investment-1)
                    ac1.append(account1); ac2.append(account2)
                    money.append(total)
                    trade_pnl.append(pnl_asset1+pnl_asset2)
                    print("ROI = ", roi[-1])
                    print("trade closed\n")
                    trade = False
 
                    #开始一个新的交易
                    if(open_asset1 == "SHORT"): #若交易所1卖出
                        open_p1 = p1; open_p2 = p2
                        open_asset1 = asset1; open_asset2 = asset2
                    else:
                        open_p1 = p1; open_p2 = p2
                        open_asset1 = asset1; open_asset2 = asset2
                    new_trade = False
                    trade = True
                    print("new trade opened:", asset1, asset2, \
                          open_p1, open_p2)
 
            else:   #[if-B:if-B1]
                print("   ",df1.close.iloc[i], df2.close.iloc[i], \
                      asset1, asset2)

    return ac1,ac2,money,roi

#==============================================================================
        
#定义函数：绘图每个账户的头寸
def eval_Position(market1,market2,investment,ac1,ac2,money):
    
    
    ymax=(round(max(money)/1000)+1)*1000
    plt.figure(figsize=(18,9))
    #market1的账户ac1变化图
    plt.subplot(2,3,1)
    plt.plot(ac1)
    plt.title(market1 + "市场的资金账户余额")
    plt.xlabel("交易序列"); plt.ylabel("账户余额")
    plt.grid()
    plt.xlim([0, len(money)]); plt.ylim([0, ymax])    
    #market2的账户ac2变化图
    plt.subplot(2,3,2)
    plt.plot(ac2)
    plt.title(market2 + "市场的资金账户余额")
    plt.xlabel("交易序列"); plt.ylabel("账户余额")
    plt.grid()
    plt.xlim([0, len(money)]); plt.ylim([0, ymax])    
    #总的账户money变化图
    plt.subplot(2,3,3)
    plt.plot(np.array(money))
    plt.title("资金总额")
    plt.xlabel("交易序列"); plt.ylabel("金额")
    plt.grid()
    plt.xlim([0, len(money)]); plt.ylim([investment, ymax])   
    
#定义函数：绘图投资总收益ROI        
def eval_Roi(fsym,tsym,market1,market2,roi):
    
    
    ttltrdnum=len(roi)
    ymax=round(10*roi[-1])*10+10
    product=fsym+"/"+tsym+"："
    mktpair="市场配对("+market1+"，"+market2+")"
    plt.figure(figsize=(8,5))
    plt.plot(np.array(roi)*100, 'r')
    plt.xlabel("交易序列(总计"+str(ttltrdnum)+"次)")
    plt.ylabel("收益率%")
    plt.title(product+mktpair+", ROI = %s%%" % str(round(100*roi[-1],2)))
    plt.xlim([0, len(roi)]); plt.ylim([0, ymax])
    plt.grid()

#定义函数：选择价差均值avg最大的2个市场。风险偏好：进取型
def select2Markets_TopSpreadAvg(fsym,tsym,begdate,enddate):
    #抓取所选产品最活跃的交易所Top5，放入市场列表markets中
    markets=fetchCrypto_Exchange(fsym,tsym,5)

    #基于投资风险偏好，选择2个价差最大的交易所
    #抓取指定交易所列表markets中所选产品在指定期间内的收盘价，放入收盘价cp中
    cp=fetchCrypto_Price_byExchList(fsym,tsym,markets,begdate,enddate) 

    #计算cp中任意2个市场组合之间所选产品的价差估计
    dist1,dist2=calcSpread_in2Markets(cp)
    
    #显示任意2个市场组合收盘价价差的均值和标准差
    print("\n*****Sorted by descending average of market spread\n",dist1)
    
    market1=dist1.loc[0,'Market1']; market2=dist1.loc[0,'Market2']
    return market1,market2

#定义函数：选择价差标准差std最小的2个市场。风险偏好：保守型
def select2Markets_BottomSpreadStd(fsym,tsym,begdate,enddate):
    #抓取所选产品最活跃的交易所Top5，放入市场列表markets中
    markets=fetchCrypto_Exchange(fsym,tsym,5)

    #基于投资风险偏好，选择2个价差最大的交易所
    #抓取指定交易所列表markets中所选产品在指定期间内的收盘价，放入收盘价cp中
    cp=fetchCrypto_Price_byExchList(fsym,tsym,markets,begdate,enddate) 

    #计算cp中任意2个市场组合之间所选产品的价差估计
    dist1,dist2=calcSpread_in2Markets(cp)
    dist2=dist2.reset_index(drop = True)    #重新索引
    #显示任意2个市场组合收盘价价差的均值和标准差
    print("\n*****Sorted by descending std of market spread\n",dist2)
    
    market1=dist2.loc[len(dist2)-1,'Market1']
    market2=dist2.loc[len(dist2)-1,'Market2']
    return market1,market2   

#定义函数：选择市场后，一步实现所有MSA操作
def implementMSA_Strategy(fsym,tsym,investment,account1,account2,position, \
                          market1,market2,begdate,enddate):
   df1,df2=evalSpread_in2Markets(fsym,tsym,market1,market2,begdate,enddate)
   ac1,ac2,money,roi=backtestMSA_Strategy( \
             investment,account1,account2,position,df1,df2)
   #省略此步骤：eval_Position(investment,ac1,ac2,money)
   eval_Roi(fsym,tsym,market1,market2,roi) 
   return roi
#==============================================================================



# -*- coding: utf-8 -*-

# 绝对引用指定目录中的模块
import sys
sys.path.insert(0,r'S:\siat\siat')
from option_china import *
#==============================================================================
#查找中国商品期权的常见品种
df1=option_comm_china()

#查找中国黄金期权的可用合约
df2=option_comm_china('黄金')

#查找中国黄金期权au2112和au2202的具体合约（看涨/看跌合约）
df3=option_comm_china('黄金','au2112')
df3b=option_comm_china('黄金','au2202')
#同一时刻行权价对期权合约的影响：行权价越高，看涨期权合约价格越低，看跌期权合约价越高

#单一期权合约价格的运动趋势：默认不带趋势线
df4=option_comm_trend_china('au2202C364','2021-9-1','2021-9-30',power=4)

#期权方向对合约价格走势的影响：看涨/看跌期权合约的价格走势正好相反
df5=option_comm_trend_china(['au2202C364','au2202P364'],'2021-8-1','2021-9-30')

#行权价对合约价格时间序列走势的影响：看涨期权，行权价低者合约价高
df6=option_comm_trend_china(['au2112C328','au2112C364'],'2021-8-1','2021-9-30')
#行权价对合约价格时间序列走势的影响：看跌期权，行权价低者合约价低，与看涨期权正好相反
df6b=option_comm_trend_china(['au2112P328','au2112P364'],'2021-8-1','2021-9-30')

#到期时间对合约价格走势的影响：看涨期权，到期日近者合约价低（时间价值少）
df7=option_comm_trend_china(['au2112C364','au2202C364'],'2021-8-1','2021-9-30')
#到期时间对合约价格走势的影响：看跌期权，到期日近者合约价低，与看涨期权一致
df7b=option_comm_trend_china(['au2112P364','au2202P364'],'2021-8-1','2021-9-30')



#==============================================================================



# -*- coding: utf-8 -*-
"""
本模块功能：债券，基础层函数
所属工具包：证券投资分析工具SIAT 
SIAT：Security Investment Analysis Tool
创建日期：2020年1月8日
最新修订日期：2020年5月10日
作者：王德宏 (WANG Dehong, Peter)
作者单位：北京外国语大学国际商学院
版权所有：王德宏
用途限制：仅限研究与教学使用，不可商用！商用需要额外授权。
特别声明：作者不对使用本工具进行证券投资导致的任何损益负责！
"""

#==============================================================================
#关闭所有警告
import warnings; warnings.filterwarnings('ignore')
from siat.common import *
from siat.translate import *
#==============================================================================
def macD0(c,y,F,n):
    """
    功能：计算债券的麦考莱久期期数
    输入参数：
    1、每期票面利率c
    2、每期到期收益率y，市场利率，贴现率
    3、票面价值F
    3、到期期数n
    输出：麦考莱久期的期数（不一定是年数）
    """
    #生成期数序列
    import pandas as pd
    t=pd.Series(range(1,(n+1)))
    
    #计算未来票息和面值的现值
    p=sum(c*F/(1+y)**t)+F/(1+y)**len(t)
    #计算未来票息和面值的加权现值
    wp=sum(c*F*t/(1+y)**t)+F*len(t)/(1+y)**len(t)
    
    return wp/p    

if __name__=='__main__':
    c=0.08/2; y=0.1/2; n=6; F=100
    print(macD0(c,y,F,n))
#==============================================================================
def macD(cr,ytm,nyears,ctimes=2,fv=100):
    """
    功能：计算债券的麦考莱久期年数
    输入参数：
    1、年票面利率cr
    2、年到期收益率ytm，市场利率，贴现率
    3、到期年数nyears
    4、每年付息次数ctimes
    5、票面价值fv
    输出：麦考莱久期（年数）
    """
    
    #转换为每期付息的参数
    c=cr/ctimes; y=ytm/ctimes; F=fv; n=nyears*ctimes
    
    #计算麦考莱久期期数
    d=macD0(c,y,F,n)
    #转换为麦考莱久期年数：年数=期数/每年付息次数
    D=round(d/ctimes,2)
    
    return D                    

if __name__=='__main__':
    cr=0.08; ytm=0.1; nyears=3; ctimes=2; fv=100
    print(macD(cr,ytm,nyears))

#==============================================================================
def MD0(c,y,F,n):
    """
    功能：计算债券的修正久期期数
    输入参数：
    1、每期票面利率c
    2、每期到期收益率y，市场利率，贴现率
    3、票面价值F
    4、到期期数n
    输出：修正久期期数（不一定是年数）
    """
    #修正麦考莱久期
    md=macD0(c,y,F,n)/(1+y)
    
    return md

if __name__=='__main__':
    c=0.08/2; y=0.1/2; n=6; F=100
    print(MD0(c,y,F,n))    
    
#==============================================================================
def MD(cr,ytm,nyears,ctimes=2,fv=100):
    """
    功能：计算债券的修正久期年数
    输入参数：
    1、年票面利率cr
    2、年到期收益率ytm，市场利率，贴现率
    3、到期年数nyears
    4、每年付息次数ctimes
    5、票面价值fv
    输出：修正久期（年数）
    """
    
    #转换为每期付息的参数
    c=cr/ctimes; y=ytm/ctimes; F=fv; n=nyears*ctimes
    
    #计算久期期数
    d=MD0(c,y,F,n)
    #转换为久期年数：年数=期数/每年付息次数
    D=round(d/ctimes,2)
    
    return D                    

if __name__=='__main__':
    cr=0.08; ytm=0.1; nyears=3; ctimes=2; fv=100
    print(MD(cr,ytm,nyears))


    
#==============================================================================
def DD0(c,y,F,n):
    """
    功能：计算债券的美元久期期数
    输入参数：
    1、每期票面利率c
    2、每期到期收益率y，市场利率，贴现率
    3、票面价值F
    4、到期期数n
    输出：美元久期期数（不一定是年数）
    """
    #生成期数序列
    import pandas as pd
    t=pd.Series(range(1,(n+1)))
    
    #计算现值
    p=sum(c*F/(1+y)**t)+F/(1+y)**len(t)
    #美元久期期数
    dd=MD0(c,y,F,n)*p
    
    return dd

if __name__=='__main__':
    c=0.08/2; y=0.1/2; n=6; F=100
    print(DD0(c,y,F,n))    
    
#==============================================================================    
def DD(cr,ytm,nyears,ctimes=2,fv=100):
    """
    功能：计算债券的美元久期年数
    输入参数：
    1、年票面利率cr
    2、年到期收益率ytm，市场利率，贴现率
    3、到期年数nyears
    4、每年付息次数ctimes
    5、票面价值fv
    输出：美元久期（年数）
    """
    
    #转换为每期付息的参数
    c=cr/ctimes; y=ytm/ctimes; F=fv; n=nyears*ctimes
    
    #计算久期期数
    d=DD0(c,y,F,n)
    #转换为久期年数：年数=期数/每年付息次数
    D=round(d/ctimes,2)
    
    return D                    

if __name__=='__main__':
    cr=0.08; ytm=0.1; nyears=3; ctimes=2; fv=100
    print(DD(cr,ytm,nyears))

#==============================================================================    
def ED0(c,y,F,n,per):
    """
    功能：计算债券的有效久期期数
    输入参数：
    1、每期票面利率c
    2、每期到期收益率y，市场利率，贴现率
    3、票面价值F
    4、到期期数n
    5、到期收益率的变化幅度，1个基点=0.01%=0.0001
    输出：有效久期期数（不一定是年数）
    """
    #生成期数序列
    import pandas as pd
    t=pd.Series(range(1,(n+1)))
    
    #计算到期收益率变化前的现值
    p0=sum(c*F/(1+y)**t)+F/(1+y)**len(t)
    #计算到期收益率增加一定幅度后的现值
    p1=sum(c*F/(1+y+per)**t)+F/(1+y+per)**len(t)
    #计算到期收益率减少同等幅度后的现值
    p2=sum(c*F/(1+y-per)**t)+F/(1+y-per)**len(t)
    #久期期数
    ed=(p2-p1)/(2*p0*per)
    
    return ed

if __name__=='__main__':
    c=0.08/2; y=0.1/2; n=6; F=100; per=0.001/2
    print(ED0(c,y,F,n,per))    
    
#==============================================================================    
def ED(cr,ytm,nyears,peryear,ctimes=2,fv=100):
    """
    功能：计算债券的有效久期年数
    输入参数：
    1、年票面利率cr
    2、年到期收益率ytm，市场利率，贴现率
    3、到期年数nyears
    4、年到期收益率变化幅度peryear
    5、每年付息次数ctimes
    6、票面价值fv
    输出：有效久期（年数）
    """
    
    #转换为每期付息的参数
    c=cr/ctimes; y=ytm/ctimes; per=peryear/ctimes; F=fv; n=nyears*ctimes
    
    #计算久期期数
    d=ED0(c,y,F,n,per)
    #转换为久期年数：年数=期数/每年付息次数
    D=round(d/ctimes,2)
    
    return D                    

if __name__=='__main__':
    cr=0.08; ytm=0.1; nyears=3; ctimes=2; fv=100; peryear=0.001
    print(ED(cr,ytm,nyears,peryear))
    
    cr=0.095; ytm=0.1144; nyears=8; ctimes=2; fv=1000; peryear=0.0005
    print(ED(cr,ytm,nyears,peryear))    
#==============================================================================    
def CFD0(c,y,F,n):
    """
    功能：计算债券的封闭式久期期数
    输入参数：
    1、每期票面利率c
    2、每期到期收益率y，市场利率，贴现率
    3、票面价值F
    4、到期期数n
    输出：久期期数（不一定是年数）
    """
    #生成期数序列
    import pandas as pd
    t=pd.Series(range(1,(n+1)))
    
    #计算到期收益率变化前的现值
    p=sum(c*F/(1+y)**t)+F/(1+y)**len(t)
    
    #计算分子第1项
    nm1=(c*F) * ((1+y)**(n+1)-(1+y)-y*n) / ((y**2)*((1+y)**n))
    #计算分子第2项
    nm2=F*(n/((1+y)**n))
    
    #计算封闭式久期
    cfd=(nm1+nm2)/p
    
    return cfd

if __name__=='__main__':
    c=0.095/2; y=0.1144/2; n=16; F=1000
    print(CFD0(c,y,F,n))    
#==============================================================================    
def CFD(cr,ytm,nyears,ctimes=2,fv=100):
    """
    功能：计算债券的封闭式年数
    输入参数：
    1、年票面利率cr
    2、年到期收益率ytm，市场利率，贴现率
    3、到期年数nyears
    4、每年付息次数ctimes
    5、票面价值fv
    输出：久期（年数）
    """
    
    #转换为每期付息的参数
    c=cr/ctimes; y=ytm/ctimes; F=fv; n=nyears*ctimes
    
    #计算久期期数
    d=CFD0(c,y,F,n)
    #转换为久期年数：年数=期数/每年付息次数
    cfd=round(d/ctimes,2)
    
    return cfd                    

if __name__=='__main__':
    cr=0.08; ytm=0.1; nyears=3; ctimes=2; fv=100
    print(CFD(cr,ytm,nyears))
#==============================================================================    
def C0(c,y,F,n):
    """
    功能：计算债券的凸度期数
    输入参数：
    1、每期票面利率c
    2、每期到期收益率y，市场利率，贴现率
    3、票面价值F
    4、到期期数n
    输出：到期收益率变化幅度为per时债券价格的变化幅度
    """    
    #生成期数序列
    import pandas as pd
    t=pd.Series(range(1,(n+1)))
    
    #计算未来现金流的现值
    p=sum(c*F/(1+y)**t)+F/(1+y)**len(t)
    #计算未来现金流的加权现值：权重为第t期的(t**2+t)
    w2p=sum(c*F*(t**2+t)/(1+y)**t)+F*(len(t)**2+len(t))/(1+y)**len(t)
    #计算凸度
    c0=w2p/(p*(1+y)**2)
    
    return c0

if __name__=='__main__':
    c=0.08/2; y=0.1/2; n=6; F=100
    print(C0(c,y,F,n)) 
#==============================================================================    
def convexity(cr,ytm,nyears,ctimes=2,fv=100):
    """
    功能：计算债券的凸度年数
    输入参数：
    1、年票面利率cr
    2、年到期收益率ytm，市场利率，贴现率
    3、到期年数nyears
    4、每年付息次数ctimes
    5、票面价值fv
    输出：凸度（年数）
    """
    #转换为每期付息的参数
    c=cr/ctimes; y=ytm/ctimes; F=fv; n=nyears*ctimes
    
    #计算凸度期数
    c=C0(c,y,F,n)
    #转换为凸度年数：年数=期数/每年付息次数**2
    cyears=round(c/ctimes**2,2)
    
    return cyears                    

if __name__=='__main__':
    cr=0.08; ytm=0.1; nyears=3; ctimes=2; fv=100
    print(convexity(cr,ytm,nyears))
#==============================================================================    
def ytm_risk(cr,ytm,nyears,peryear,ctimes=2,fv=100):
    """
    功能：计算债券的利率风险，即市场利率（到期收益率）变动将带来的债券价格变化率
    输入参数：
    1、年票面利率cr
    2、年到期收益率ytm，市场利率，贴现率
    3、到期年数nyears
    4、每期到期收益率（市场利率）变化的幅度per，100个基点=1%
    5、每年付息次数ctimes
    6、票面价值fv
    输出：到期收益率变化幅度导致的债券价格变化率
    """
    #转换为每期付息的参数
    c=cr/ctimes; y=ytm/ctimes; F=fv; n=nyears*ctimes
    
    #计算到期收益率变化对债券价格的影响：第1部分
    b0=-MD0(c,y,F,n)/2*peryear
    #计算到期收益率变化对债券价格的影响：第2部分
    b1=(0.5*C0(c,y,F,n)/ctimes**2)*peryear**2
    #债券价格的变化率
    p_pct=round(b0+b1,4)
        
    return p_pct

if __name__=='__main__':
    cr=0.08; ytm=0.1; nyears=3; peryear=0.01
    print(ytm_risk(cr,ytm,nyears,peryear))        

#==============================================================================    
#==============================================================================    
def interbank_bond_issue_detail(fromdate,todate):
    """
    功能：获得银行间债券市场发行明细
    输入：开始日期fromdate，截止日期todate
    """
    #检查期间的合理性
    result,start,end=check_period(fromdate, todate)
    if result is None:
        print("...Error(interbank_bond_issue_detail), invalid period:",fromdate,todate)
        return None
    
    #####银行间市场的债券发行数据
    import akshare as ak
    #获得债券发行信息第1页
    print("\n...Searching for bond issuance: ",end='')
    bond_issue=ak.get_bond_bank(page_num=1)    

    import pandas as pd
    from datetime import datetime
    #获得债券发行信息后续页
    maxpage=999
    for pn in range(2,maxpage):
        print_progress_bar(pn,2,maxpage)
        try:
            #防止中间一次失败导致整个过程失败
            bi=ak.get_bond_bank(page_num=pn)
            bond_issue=bond_issue.append(bi)
        except:
            #后续的网页已经变得无法抓取
            print("...Unexpected get_bond_bank(interbank_bond_issue_detail), page_num",pn)
            break        
        
        #判断是否超过了开始日期
        bistartdate=bi.tail(1)['releaseTime'].values[0]
        bistartdate2=pd.to_datetime(bistartdate)
        if bistartdate2 < start: break
    print(" Done!")        
    
    #删除重复项，按日期排序
    bond_issue.drop_duplicates(keep='first',inplace=True)
    bond_issue.sort_values(by=['releaseTime'],ascending=[False],inplace=True)    
    #转换日期项
    lway1=lambda x: datetime.strptime(x,'%Y-%m-%d %H:%M:%S')
    bond_issue['releaseTime2']=bond_issue['releaseTime'].apply(lway1)    
    
    #提取年月日信息
    lway2=lambda x: x.year
    bond_issue['releaseYear']=bond_issue['releaseTime2'].map(lway2).astype('str')
    lway3=lambda x: x.month
    bond_issue['releaseMonth']=bond_issue['releaseTime2'].map(lway3).astype('str')
    lway4=lambda x: x.day
    bond_issue['releaseDay']=bond_issue['releaseTime2'].map(lway4).astype('str')
    lway5=lambda x: x.weekday() + 1
    bond_issue['releaseWeekDay']=bond_issue['releaseTime2'].map(lway5).astype('str')
    lway6=lambda x: x.date()
    bond_issue['releaseDate']=bond_issue['releaseTime2'].map(lway6).astype('str')
    
    #过滤日期
    bond_issue=bond_issue.reset_index(drop = True)
    bond_issue1=bond_issue.drop(bond_issue[bond_issue['releaseTime2']<start].index)
    bond_issue1=bond_issue1.reset_index(drop = True)
    bond_issue2=bond_issue1.drop(bond_issue1[bond_issue1['releaseTime2']>end].index)
    bond_issue2=bond_issue2.reset_index(drop = True)
    #转换字符串到金额
    bond_issue2['issueAmount']=bond_issue2['firstIssueAmount'].astype('float64')
    
    return bond_issue2
    
if __name__=='__main__':
    fromdate='2020-4-25'    
    todate='2020-4-28'
    ibbi=interbank_bond_issue_detail(fromdate,todate)
    
#==============================================================================
